import sys
import platform

__all__ = ['PY3', 'Semaphore']

PY3 = sys.version_info[0] >= 3
IS_PYPY = platform.python_implementation().lower() == 'pypy'


def get_bytes(value):
    if hasattr(value, 'encode'):
        try:
            value = value.encode('utf-8')
        except UnicodeError:
            # if we can't encode the value just pass it along
            pass
    return value


def get_string(value):
    if hasattr(value, 'decode'):
        try:
            value = value.decode('utf-8')
        except UnicodeError:
            # if we can't decode the value just pass it along
            pass
    else:
        value = str(value)
    return value

if PY3:
    from threading import Semaphore
    from queue import Queue, Empty  # noqa
    range = range

    def iteritems(d, **kw):
        return iter(d.items(**kw))

    def itervalues(d, **kw):
        return iter(d.values(**kw))

    def iterkeys(d, **kw):
        return iter(d.keys(**kw))

    buffer = memoryview

else:
    range = xrange
    from threading import Condition, Lock
    # could use monotonic.monotonic() backport as well here...
    from time import time as _time
    from Queue import Queue, Empty  # noqa

    def iteritems(d, **kw):
        return d.iteritems(**kw)

    def itervalues(d, **kw):
        return d.itervalues(**kw)

    def iterkeys(d, **kw):
        return d.iterkeys(**kw)

    buffer = buffer

    # -- begin unmodified backport of threading.Semaphore from Python 3.4 -- #
    class Semaphore(object):
        """This class implements semaphore objects.

        Semaphores manage a counter representing the number of release() calls minus
        the number of acquire() calls, plus an initial value. The acquire() method
        blocks if necessary until it can return without making the counter
        negative. If not given, value defaults to 1.

        Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011,
        2012, 2013, 2014, 2015 Python Software Foundation.  All rights reserved.
        """

        # After Tim Peters' semaphore class, but not quite the same (no maximum)

        def __init__(self, value=1):
            if value < 0:
                raise ValueError("semaphore initial value must be >= 0")
            self._cond = Condition(Lock())
            self._value = value

        def acquire(self, blocking=True, timeout=None):
            """Acquire a semaphore, decrementing the internal counter by one.

            When invoked without arguments: if the internal counter is larger than
            zero on entry, decrement it by one and return immediately. If it is zero
            on entry, block, waiting until some other thread has called release() to
            make it larger than zero. This is done with proper interlocking so that
            if multiple acquire() calls are blocked, release() will wake exactly one
            of them up. The implementation may pick one at random, so the order in
            which blocked threads are awakened should not be relied on. There is no
            return value in this case.

            When invoked with blocking set to true, do the same thing as when called
            without arguments, and return true.

            When invoked with blocking set to false, do not block. If a call without
            an argument would block, return false immediately; otherwise, do the
            same thing as when called without arguments, and return true.

            When invoked with a timeout other than None, it will block for at
            most timeout seconds.  If acquire does not complete successfully in
            that interval, return false.  Return true otherwise.

            """
            if not blocking and timeout is not None:
                raise ValueError("can't specify timeout for non-blocking acquire")
            rc = False
            endtime = None
            with self._cond:
                while self._value == 0:
                    if not blocking:
                        break
                    if timeout is not None:
                        if endtime is None:
                            endtime = _time() + timeout
                        else:
                            timeout = endtime - _time()
                            if timeout <= 0:
                                break
                    self._cond.wait(timeout)
                else:
                    self._value -= 1
                    rc = True
            return rc

        __enter__ = acquire

        def release(self):
            """Release a semaphore, incrementing the internal counter by one.

            When the counter is zero on entry and another thread is waiting for it
            to become larger than zero again, wake up that thread.

            """
            with self._cond:
                self._value += 1
                self._cond.notify()

        def __exit__(self, t, v, tb):
            self.release()
    # -- end backport of Semaphore from Python 3.4 -- #



import unittest2

from hashlib import sha1
from pykafka.partitioners import GroupHashingPartitioner


class TestGroupHashingPartitioner(unittest2.TestCase):

    def test_valid_inputs_success(self):
        # Test data; 1st element is group size, 2nd is total partition count, 3rd is the key
        key = 'foo'.encode('utf-8')
        data = [[1, 16, key],
                [2, 16, key],
                [4, 16, key],
                [16, 16, key],
                [1, 1, key]]
        for row in data:
            self._run_test(row[0], row[1], row[2])

    def test_invalid_inputs_error(self):
        key = 'foo'.encode('utf-8')
        data = [[0, 16, key],
                [17, 16, key]]
        for row in data:
            with self.assertRaises(ValueError):
                self._run_test(row[0], row[1], row[2])
                continue

    def test_create_with_zero_group_size_raises_error(self):
        with self.assertRaises(ValueError):
            GroupHashingPartitioner(hash_func=None, group_size=0)

    def test_create_with_negative_group_size_raises_error(self):
        with self.assertRaises(ValueError):
            GroupHashingPartitioner(hash_func=None, group_size=-1)

    def test_missing_hash_function_raises_error(self):
        with self.assertRaises(ValueError):
            GroupHashingPartitioner(hash_func=None, group_size=1)

    def _run_test(self, group_size, total_partition_count, key):
        def hash_func(k): return int(sha1(k).hexdigest(), 16)
        # Instead of one hash for each key, generate a list of n possible hashes, where n=group_size
        hashed_keys = [abs(hash_func(key) + x) for x in range(group_size)]
        # Obtain a list of valid partitions
        valid_partitions = list(map(lambda x: x % total_partition_count, hashed_keys))
        self.assertEquals(len(valid_partitions), group_size)

        # Call the partitioner and check that the returned partition is in the valid list
        partitioner = GroupHashingPartitioner(hash_func, group_size)
        x = partitioner.__call__(list(range(total_partition_count)), key)
        self.assertTrue(x in valid_partitions)



"""
Author: Keith Bourgoin, Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class KafkaException(Exception):
    """Generic exception type. The base of all pykafka exception types."""
    pass


class UnicodeException(Exception):
    """Indicates that an error was encountered while processing a unicode string"""
    pass


class NoBrokersAvailableError(KafkaException):
    """Indicates that no brokers were available to the cluster's metadata update attempts
    """
    pass


class LeaderNotFoundError(KafkaException):
    """Indicates that the leader broker for a given partition was not found during
        an update in response to a MetadataRequest
    """
    pass


class SocketDisconnectedError(KafkaException):
    """Indicates that the socket connecting this client to a kafka broker has
        become disconnected
    """
    pass


class ProduceFailureError(KafkaException):
    """Indicates a generic failure in the producer"""
    pass


class ConsumerStoppedException(KafkaException):
    """Indicates that the consumer was stopped when an operation was attempted that required it to be running"""
    pass


class NoMessagesConsumedError(KafkaException):
    """Indicates that no messages were returned from a MessageSet"""
    pass


class MessageSetDecodeFailure(KafkaException):
    """Indicates a generic failure in the decoding of a MessageSet from the broker"""
    pass


class ProducerQueueFullError(KafkaException):
    """Indicates that one or more of the AsyncProducer's internal queues contain at least max_queued_messages messages"""
    pass


class ProducerStoppedException(KafkaException):
    """Raised when the Producer is used while not running"""
    pass


class OffsetRequestFailedError(KafkaException):
    """Indicates that OffsetRequests for offset resetting failed more times than the configured maximum"""
    pass


class PartitionOwnedError(KafkaException):
    """Indicates a given partition is still owned in Zookeeper."""

    def __init__(self, partition, *args, **kwargs):
        super(PartitionOwnedError, self).__init__(*args, **kwargs)
        self.partition = partition


"""
Protocol Client Exceptions
https://cwiki.apache.org/confluence/display/KAFKA/A+Guide+To+The+Kafka+Protocol#AGuideToTheKafkaProtocol-ErrorCodes

NOTE: Don't raise these from client code unless it's in direct response to an error
code from the broker. When that's not the case, the exception raised should instead be
a subclass of KafkaException.
"""


class ProtocolClientError(KafkaException):
    """Base class for protocol errors"""
    ERROR_CODE = None


class UnknownError(ProtocolClientError):
    """An unexpected server error"""
    ERROR_CODE = -1


class OffsetOutOfRangeError(ProtocolClientError):
    """The requested offset is outside the range of offsets maintained by the
        server for the given topic/partition.
    """
    ERROR_CODE = 1


class InvalidMessageError(ProtocolClientError):
    """This indicates that a message contents does not match its CRC"""
    ERROR_CODE = 2


class UnknownTopicOrPartition(ProtocolClientError):
    """This request is for a topic or partition that does not exist on this
        broker.
    """
    ERROR_CODE = 3


class InvalidMessageSize(ProtocolClientError):
    """The message has a negative size"""
    ERROR_CODE = 4


class LeaderNotAvailable(ProtocolClientError):
    """This error is thrown if we are in the middle of a leadership election
        and there is currently no leader for this partition and hence it is
        unavailable for writes.
    """
    ERROR_CODE = 5


class NotLeaderForPartition(ProtocolClientError):
    """This error is thrown if the client attempts to send messages to a
        replica that is not the leader for some partition. It indicates that
        the client's metadata is out of date.
    """
    ERROR_CODE = 6


class RequestTimedOut(ProtocolClientError):
    """This error is thrown if the request exceeds the user-specified time
        limit in the request.
    """
    ERROR_CODE = 7


class MessageSizeTooLarge(ProtocolClientError):
    """The server has a configurable maximum message size to avoid unbounded
        memory allocation. This error is thrown if the client attempts to
        produce a message larger than this maximum.
    """
    ERROR_CODE = 10


class OffsetMetadataTooLarge(ProtocolClientError):
    """If you specify a string larger than configured maximum for offset
        metadata
    """
    ERROR_CODE = 12


class GroupLoadInProgress(ProtocolClientError):
    """The broker returns this error code for an offset fetch request if it is
        still loading offsets (after a leader change for that offsets topic
        partition), or in response to group membership requests (such as
        heartbeats) when group metadata is being loaded by the coordinator.
    """
    ERROR_CODE = 14


class GroupCoordinatorNotAvailable(ProtocolClientError):
    """The broker returns this error code for consumer metadata requests or
        offset commit requests if the offsets topic has not yet been created.
    """
    ERROR_CODE = 15


class NotCoordinatorForGroup(ProtocolClientError):
    """The broker returns this error code if it receives an offset fetch or
        commit request for a consumer group that it is not a coordinator for.
    """
    ERROR_CODE = 16


class InvalidTopic(ProtocolClientError):
    """For a request which attempts to access an invalid topic (e.g. one which has
        an illegal name), or if an attempt is made to write to an internal topic
        (such as the consumer offsets topic).
    """
    ERROR_CODE = 17


class IllegalGeneration(ProtocolClientError):
    """Returned from group membership requests (such as heartbeats) when the generation
        id provided in the request is not the current generation
    """
    ERROR_CODE = 22


class InconsistentGroupProtocol(ProtocolClientError):
    """Returned in join group when the member provides a protocol type or set of protocols
        which is not compatible with the current group.
    """
    ERROR_CODE = 23


class UnknownMemberId(ProtocolClientError):
    """Returned from group requests (offset commits/fetches, heartbeats, etc) when the
        memberId is not in the current generation. Also returned if SimpleConsumer is
        incorrectly instantiated with a non-default consumer_id.
    """
    ERROR_CODE = 25


class InvalidSessionTimeout(ProtocolClientError):
    """Returned in join group when the requested session timeout is outside of the allowed
        range on the broker
    """
    ERROR_CODE = 26


class RebalanceInProgress(ProtocolClientError):
    """Returned in heartbeat requests when the coordinator has begun rebalancing the
        group. This indicates to the client that it should rejoin the group.
    """
    ERROR_CODE = 27


class TopicAuthorizationFailed(ProtocolClientError):
    """Returned by the broker when the client is not authorized to access the requested
        topic.
    """
    ERROR_CODE = 29


class GroupAuthorizationFailed(ProtocolClientError):
    """Returned by the broker when the client is not authorized to access a particular
    groupId.
    """
    ERROR_CODE = 30


ERROR_CODES = dict(
    (exc.ERROR_CODE, exc)
    for exc in (UnknownError,
                OffsetOutOfRangeError,
                InvalidMessageError,
                UnknownTopicOrPartition,
                InvalidMessageSize,
                LeaderNotAvailable,
                NotLeaderForPartition,
                RequestTimedOut,
                MessageSizeTooLarge,
                OffsetMetadataTooLarge,
                GroupLoadInProgress,
                GroupCoordinatorNotAvailable,
                NotCoordinatorForGroup,
                InvalidTopic,
                IllegalGeneration,
                InconsistentGroupProtocol,
                UnknownMemberId,
                InvalidSessionTimeout,
                RebalanceInProgress,
                TopicAuthorizationFailed,
                GroupAuthorizationFailed)
)


class RdKafkaException(KafkaException):
    """Error in rdkafka extension that hasn't any equivalent pykafka exception

    In `pykafka.rdkafka._rd_kafka` we try hard to emit the same exceptions
    that the pure pykafka classes emit.  This is a fallback for the few cases
    where we can't find a suitable exception
    """
    pass


class RdKafkaStoppedException(RdKafkaException):
    """Consumer or producer handle was stopped

    Raised by the C extension, to be translated to ConsumerStoppedException or
    ProducerStoppedException by the caller
    """
    pass



# - coding: utf-8 -
import struct

from .base import Request, Response
from ..utils import struct_helpers


class ConsumerGroupProtocolMetadata(object):
    """
    Protocol specification::
    ProtocolMetadata => Version Subscription UserData
        Version => int16
        Subscription => [Topic]
            Topic => string
        UserData => bytes
    """
    def __init__(self, version=0, topic_names=None, user_data=b"testuserdata"):
        self.version = version
        self.topic_names = topic_names or [b"dummytopic"]
        self.user_data = user_data

    def __len__(self):
        # version + len(topic names)
        size = 2 + 4
        for topic_name in self.topic_names:
            # len(topic_name) + topic_name
            size += 2 + len(topic_name)
        # len(user data) + user data
        size += 4 + len(self.user_data)
        return size

    def get_bytes(self):
        output = bytearray(len(self))
        offset = 0
        fmt = '!hi'
        struct.pack_into(fmt, output, offset, self.version, len(self.topic_names))
        offset += struct.calcsize(fmt)
        for topic_name in self.topic_names:
            fmt = '!h%ds' % len(topic_name)
            struct.pack_into(fmt, output, offset, len(topic_name), topic_name)
            offset += struct.calcsize(fmt)
        fmt = '!i%ds' % len(self.user_data)
        struct.pack_into(fmt, output, offset, len(self.user_data), self.user_data)
        offset += struct.calcsize(fmt)
        return output

    @classmethod
    def from_bytestring(cls, buff):
        if len(buff) == 0:
            return cls()
        fmt = 'h [S] Y'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        version = response[0]
        topic_names = response[1]
        user_data = response[2]
        return cls(version, topic_names, user_data)


class JoinGroupRequest(Request):
    """A group join request
    Specification::
    JoinGroupRequest => GroupId SessionTimeout MemberId ProtocolType GroupProtocols
        GroupId => string
        SessionTimeout => int32
        MemberId => string
        ProtocolType => string
        GroupProtocols => [ProtocolName ProtocolMetadata]
            ProtocolName => string
            ProtocolMetadata => bytes
    """
    API_KEY = 11

    def __init__(self,
                 group_id,
                 member_id,
                 topic_name,
                 membership_protocol,
                 session_timeout=30000):
        """Create a new group join request"""
        self.protocol = membership_protocol
        self.group_id = group_id
        self.session_timeout = session_timeout
        self.member_id = member_id
        self.protocol_type = self.protocol.protocol_type
        self.group_protocols = [(self.protocol.protocol_name,
                                 bytes(self.protocol.metadata.get_bytes()))]

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # Header + group id + session timeout
        size = self.HEADER_LEN + 2 + len(self.group_id) + 4
        # + member id + protocol type + len(group protocols)
        size += 2 + len(self.member_id) + 2 + len(self.protocol_type) + 4
        # metadata tuples
        for name, metadata in self.group_protocols:
            size += 2 + len(name) + 4 + len(metadata)
        return size

    def get_bytes(self):
        """Serialize the message
        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        output = bytearray(len(self))
        self._write_header(output)
        offset = self.HEADER_LEN
        fmt = '!h%dsih%dsh%dsi' % (len(self.group_id), len(self.member_id),
                                   len(self.protocol_type))
        struct.pack_into(fmt, output, offset, len(self.group_id), self.group_id,
                         self.session_timeout, len(self.member_id), self.member_id,
                         len(self.protocol_type), self.protocol_type,
                         len(self.group_protocols))
        offset += struct.calcsize(fmt)
        for protocol_name, protocol_metadata in self.group_protocols:
            fmt = '!h%dsi%ds' % (len(protocol_name), len(protocol_metadata))
            struct.pack_into(fmt, output, offset, len(protocol_name), protocol_name,
                             len(protocol_metadata), protocol_metadata)
            offset += struct.calcsize(fmt)
        return output


class JoinGroupResponse(Response):
    """A group join response
    Specification::
    JoinGroupResponse => ErrorCode GenerationId GroupProtocol LeaderId MemberId Members
        ErrorCode => int16
        GenerationId => int32
        GroupProtocol => string
        LeaderId => string
        MemberId => string
        Members => [MemberId MemberMetadata]
            MemberId => string
            MemberMetadata => bytes
    """
    def __init__(self, buff):
        """Deserialize into a new Response
        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = 'hiSSS[SY ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        self.error_code = response[0]
        self.generation_id = response[1]
        self.group_protocol = response[2]
        self.leader_id = response[3]
        self.member_id = response[4]
        self.members = {_id: ConsumerGroupProtocolMetadata.from_bytestring(meta)
                        for _id, meta in response[5]}


class MemberAssignment(object):
    """
    Protocol specification::
    MemberAssignment => Version PartitionAssignment
        Version => int16
        PartitionAssignment => [Topic [Partition]]
            Topic => string
            Partition => int32
        UserData => bytes
    """
    def __init__(self, partition_assignment, version=1):
        self.version = version
        self.partition_assignment = partition_assignment

    @classmethod
    def from_bytestring(cls, buff):
        if len(buff) == 0:
            return cls(tuple())
        fmt = 'h [S [i ] ]'
        response = struct_helpers.unpack_from(fmt, buff, 0)

        version = response[0]
        partition_assignment = response[1]
        return cls(partition_assignment, version=version)

    def __len__(self):
        # version + len(partition assignment)
        size = 2 + 4
        for topic_name, partitions in self.partition_assignment:
            # len(topic_name) + topic_name + len(partitions)
            size += 2 + len(topic_name) + 4
            size += 4 * len(partitions)
        return size

    def get_bytes(self):
        output = bytearray(len(self))
        offset = 0
        fmt = '!hi'
        struct.pack_into(fmt, output, offset, self.version,
                         len(self.partition_assignment))
        offset += struct.calcsize(fmt)
        for topic_name, partitions in self.partition_assignment:
            fmt = '!h%dsi' % len(topic_name)
            struct.pack_into(fmt, output, offset, len(topic_name), topic_name,
                             len(partitions))
            offset += struct.calcsize(fmt)
            for partition_id in partitions:
                fmt = '!i'
                struct.pack_into(fmt, output, offset, partition_id)
                offset += struct.calcsize(fmt)
        return output


class SyncGroupRequest(Request):
    """A group sync request
    Specification::
    SyncGroupRequest => GroupId GenerationId MemberId GroupAssignment
        GroupId => string
        GenerationId => int32
        MemberId => string
        GroupAssignment => [MemberId MemberAssignment]
            MemberId => string
            MemberAssignment => bytes
    """
    API_KEY = 14

    def __init__(self, group_id, generation_id, member_id, group_assignment):
        """Create a new group join request"""
        self.group_id = group_id
        self.generation_id = generation_id
        self.member_id = member_id
        self.group_assignment = group_assignment

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # Header + len(group id) + group id + generation id
        size = self.HEADER_LEN + 2 + len(self.group_id) + 4
        # + len(member id) + member id + len(group assignment)
        size += 2 + len(self.member_id) + 4
        # group assignment tuples
        for member_id, member_assignment in self.group_assignment:
            # + len(member id) + member id + len(member assignment) + member assignment
            size += 2 + len(member_id) + 4 + len(member_assignment)
        return size

    def get_bytes(self):
        """Serialize the message
        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        output = bytearray(len(self))
        self._write_header(output)
        offset = self.HEADER_LEN
        fmt = '!h%dsih%dsi' % (len(self.group_id), len(self.member_id))
        struct.pack_into(fmt, output, offset, len(self.group_id), self.group_id,
                         self.generation_id, len(self.member_id), self.member_id,
                         len(self.group_assignment))
        offset += struct.calcsize(fmt)
        for member_id, member_assignment in self.group_assignment:
            assignment_bytes = bytes(member_assignment.get_bytes())
            fmt = '!h%dsi%ds' % (len(member_id), len(assignment_bytes))
            struct.pack_into(fmt, output, offset, len(member_id), member_id,
                             len(assignment_bytes), assignment_bytes)
            offset += struct.calcsize(fmt)
        return output


class SyncGroupResponse(Response):
    """A group sync response
    Specification::
    SyncGroupResponse => ErrorCode MemberAssignment
        ErrorCode => int16
        MemberAssignment => bytes
    """
    def __init__(self, buff):
        """Deserialize into a new Response
        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = 'hY'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        self.error_code = response[0]
        self.member_assignment = MemberAssignment.from_bytestring(response[1])


class HeartbeatRequest(Request):
    """A group heartbeat request
    Specification::
    HeartbeatRequest => GroupId GenerationId MemberId
        GroupId => string
        GenerationId => int32
        MemberId => string
    """
    API_KEY = 12

    def __init__(self, group_id, generation_id, member_id):
        """Create a new heartbeat request"""
        self.group_id = group_id
        self.generation_id = generation_id
        self.member_id = member_id

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # Header + len(group id) + group id + generation id
        size = self.HEADER_LEN + 2 + len(self.group_id) + 4
        # + len(member id) + member id
        size += 2 + len(self.member_id)
        return size

    def get_bytes(self):
        """Serialize the message
        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        output = bytearray(len(self))
        self._write_header(output)
        offset = self.HEADER_LEN
        fmt = '!h%dsih%ds' % (len(self.group_id), len(self.member_id))
        struct.pack_into(fmt, output, offset, len(self.group_id), self.group_id,
                         self.generation_id, len(self.member_id), self.member_id)
        offset += struct.calcsize(fmt)
        return output


class HeartbeatResponse(Response):
    """A group heartbeat response
    Specification::
    HeartbeatResponse => ErrorCode
        ErrorCode => int16
    """
    def __init__(self, buff):
        """Deserialize into a new Response
        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = 'h'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        self.error_code = response[0]


class LeaveGroupRequest(Request):
    """A group exit request
    Specification::
    LeaveGroupRequest => GroupId MemberId
        GroupId => string
        MemberId => string
    """
    API_KEY = 13

    def __init__(self, group_id, member_id):
        """Create a new group join request"""
        self.group_id = group_id
        self.member_id = member_id

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # Header + len(group id) + group id
        size = self.HEADER_LEN + 2 + len(self.group_id)
        # + len(member id) + member id
        size += 2 + len(self.member_id)
        return size

    def get_bytes(self):
        """Serialize the message
        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        output = bytearray(len(self))
        self._write_header(output)
        offset = self.HEADER_LEN
        fmt = '!h%dsh%ds' % (len(self.group_id), len(self.member_id))
        struct.pack_into(fmt, output, offset, len(self.group_id), self.group_id,
                         len(self.member_id), self.member_id)
        offset += struct.calcsize(fmt)
        return output


class LeaveGroupResponse(Response):
    """A group exit response
    Specification::
    LeaveGroupResponse => ErrorCode
        ErrorCode => int16
    """
    def __init__(self, buff):
        """Deserialize into a new Response
        :param buff: Serialized message
        :type buff: :class:`bytearray`
        """
        fmt = 'h'
        response = struct_helpers.unpack_from(fmt, buff, 0)
        self.error_code = response[0]



import operator
import unittest2

from pykafka import protocol
from pykafka.common import CompressionType
from pykafka.membershipprotocol import RangeProtocol
from pykafka.utils.compat import buffer


class TestMetadataAPI(unittest2.TestCase):
    maxDiff = None

    def test_request(self):
        req = protocol.MetadataRequest()
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                # header
                b'\x00\x00\x00\x15'  # len(buffer)
                b'\x00\x03'  # ApiKey
                b'\x00\x00'  # api version
                b'\x00\x00\x00\x00'  # correlation id
                b'\x00\x07'  # len(client id)
                    b'pykafka'  # client id  # noqa
                # end header

                b'\x00\x00\x00\x00'  # len(topics)
            )
        )

    def test_response(self):
        cluster = protocol.MetadataResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(brokers)
                    b'\x00\x00\x00\x00'  # node id # noqa
                    b'\x00\x09'  # len(host)
                        b'localhost'  # host
                    b'\x00\x00#\x84'  # port
                b'\x00\x00\x00\x01'  # len(topic metadata)
                    b'\x00\x00'  # error code
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x02'  # len(partition metadata)
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x00'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x01'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replicas
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
            )
        )
        self.assertEqual(cluster.brokers[0].host, b'localhost')
        self.assertEqual(cluster.brokers[0].port, 9092)
        self.assertEqual(cluster.topics[b'test'].partitions[0].leader,
                         cluster.brokers[0].id)
        self.assertEqual(cluster.topics[b'test'].partitions[0].replicas,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.topics[b'test'].partitions[0].isr,
                         [cluster.brokers[0].id])

    def test_partition_error(self):
        # Response has a UnknownTopicOrPartition error for test/0
        response = protocol.MetadataResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(brokers)
                    b'\x00\x00\x00\x00'  # node is # noqa
                    b'\x00\x09'  # len(host)
                        b'localhost'  # host
                    b'\x00\x00#\x84'  # port
                b'\x00\x00\x00\x01'  # len(topic metadata)
                    b'\x00\x00'  # error code
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x02'  # len(partition metadata)
                        b'\x00\x03'  # partition error code
                            b'\x00\x00\x00\x00'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x01'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
            )
        )
        self.assertEqual(response.topics[b'test'].partitions[0].err, 3)

    def test_topic_error(self):
        # Response has a UnknownTopicOrPartition error for test/0
        response = protocol.MetadataResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(brokers)
                    b'\x00\x00\x00\x00'  # node id  # noqa
                    b'\x00\x09'  # len(host)
                        b'localhost'  # host
                    b'\x00\x00#\x84'  # port
                b'\x00\x00\x00\x01'  # len(topic metadata)
                    b'\x00\x03'  # error code
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00\x00\x00\x02'  # len(partition metadata)
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x00'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x01'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replicas
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
            )
        )
        self.assertEqual(response.topics[b'test'].err, 3)


class TestMetadataAPIV1(unittest2.TestCase):
    maxDiff = None

    def test_request(self):
        req = protocol.MetadataRequestV1()
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                # header
                b'\x00\x00\x00\x15'  # len(buffer)
                b'\x00\x03'  # ApiKey
                b'\x00\x01'  # api version
                b'\x00\x00\x00\x00'  # correlation id
                b'\x00\x07'  # len(client id)
                    b'pykafka'  # client id  # noqa
                # end header

                b'\xff\xff\xff\xff'  # len(topics)
            )
        )

    def test_response(self):
        cluster = protocol.MetadataResponseV1(
            buffer(
                b'\x00\x00\x00\x01'  # len(brokers)
                    b'\x00\x00\x00\x00'  # node id # noqa
                    b'\x00\x09'  # len(host)
                        b'localhost'  # host
                    b'\x00\x00#\x84'  # port
                    b'\xff\xff'  # len(rack)
                b'\x00\x00\x00\x00'  # controller_id
                b'\x00\x00\x00\x01'  # len(topic metadata)
                    b'\x00\x00'  # error code
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00'  # is_internal
                    b'\x00\x00\x00\x02'  # len(partition metadata)
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x00'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x01'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replicas
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
            )
        )
        self.assertEqual(cluster.brokers[0].host, b'localhost')
        self.assertEqual(cluster.brokers[0].port, 9092)
        self.assertEqual(cluster.topics[b'test'].partitions[0].leader,
                         cluster.brokers[0].id)
        self.assertEqual(cluster.topics[b'test'].partitions[0].replicas,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.topics[b'test'].partitions[0].isr,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.brokers[0].rack, None)
        self.assertEqual(cluster.controller_id, 0)
        self.assertEqual(cluster.topics[b'test'].is_internal, False)


class TestMetadataAPIV2(unittest2.TestCase):
    maxDiff = None

    def test_request(self):
        req = protocol.MetadataRequestV2()
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                # header
                b'\x00\x00\x00\x15'  # len(buffer)
                b'\x00\x03'  # ApiKey
                b'\x00\x02'  # api version
                b'\x00\x00\x00\x00'  # correlation id
                b'\x00\x07'  # len(client id)
                    b'pykafka'  # client id  # noqa
                # end header

                b'\xff\xff\xff\xff'  # len(topics)
            )
        )

    def test_response(self):
        cluster = protocol.MetadataResponseV2(
            buffer(
                b'\x00\x00\x00\x01'  # len(brokers)
                    b'\x00\x00\x00\x00'  # node id # noqa
                    b'\x00\x09'  # len(host)
                        b'localhost'  # host
                    b'\x00\x00#\x84'  # port
                    b'\xff\xff'  # len(rack)
                b'\x00\x01'  # len(cluster_id)
                    b'a'  # cluster_id
                b'\x00\x00\x00\x00'  # controller_id
                b'\x00\x00\x00\x01'  # len(topic metadata)
                    b'\x00\x00'  # error code
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00'  # is_internal
                    b'\x00\x00\x00\x02'  # len(partition metadata)
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x00'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x01'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replicas
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
            )
        )
        self.assertEqual(cluster.brokers[0].host, b'localhost')
        self.assertEqual(cluster.brokers[0].port, 9092)
        self.assertEqual(cluster.topics[b'test'].partitions[0].leader,
                         cluster.brokers[0].id)
        self.assertEqual(cluster.topics[b'test'].partitions[0].replicas,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.topics[b'test'].partitions[0].isr,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.brokers[0].rack, None)
        self.assertEqual(cluster.controller_id, 0)
        self.assertEqual(cluster.cluster_id, b"a")
        self.assertEqual(cluster.topics[b'test'].is_internal, False)


class TestMetadataAPIV3(unittest2.TestCase):
    maxDiff = None

    def test_request(self):
        req = protocol.MetadataRequestV3()
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                # header
                b'\x00\x00\x00\x15'  # len(buffer)
                b'\x00\x03'  # ApiKey
                b'\x00\x03'  # api version
                b'\x00\x00\x00\x00'  # correlation id
                b'\x00\x07'  # len(client id)
                    b'pykafka'  # client id  # noqa
                # end header

                b'\xff\xff\xff\xff'  # len(topics)
            )
        )

    def test_response(self):
        cluster = protocol.MetadataResponseV3(
            buffer(
                b'\x00\x00\x00\x00'  # throttle_time_ms
                b'\x00\x00\x00\x01'  # len(brokers)
                    b'\x00\x00\x00\x00'  # node id # noqa
                    b'\x00\x09'  # len(host)
                        b'localhost'  # host
                    b'\x00\x00#\x84'  # port
                    b'\xff\xff'  # len(rack)
                b'\x00\x01'  # len(cluster_id)
                    b'a'  # cluster_id
                b'\x00\x00\x00\x00'  # controller_id
                b'\x00\x00\x00\x01'  # len(topic metadata)
                    b'\x00\x00'  # error code
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00'  # is_internal
                    b'\x00\x00\x00\x02'  # len(partition metadata)
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x00'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x01'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replicas
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
            )
        )
        self.assertEqual(cluster.brokers[0].host, b'localhost')
        self.assertEqual(cluster.brokers[0].port, 9092)
        self.assertEqual(cluster.topics[b'test'].partitions[0].leader,
                         cluster.brokers[0].id)
        self.assertEqual(cluster.topics[b'test'].partitions[0].replicas,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.topics[b'test'].partitions[0].isr,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.brokers[0].rack, None)
        self.assertEqual(cluster.throttle_time_ms, 0)
        self.assertEqual(cluster.controller_id, 0)
        self.assertEqual(cluster.cluster_id, b"a")
        self.assertEqual(cluster.topics[b'test'].is_internal, False)


class TestMetadataAPIV4(unittest2.TestCase):
    maxDiff = None

    def test_request(self):
        req = protocol.MetadataRequestV4()
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                # header
                b'\x00\x00\x00\x16'  # len(buffer)
                b'\x00\x03'  # ApiKey
                b'\x00\x04'  # api version
                b'\x00\x00\x00\x00'  # correlation id
                b'\x00\x07'  # len(client id)
                    b'pykafka'  # client id  # noqa
                # end header

                b'\xff\xff\xff\xff'  # len(topics)
                b'\x01'  # allow_topic_autocreation
            )
        )

    def test_response(self):
        cluster = protocol.MetadataResponseV4(
            buffer(
                b'\x00\x00\x00\x00'  # throttle_time_ms
                b'\x00\x00\x00\x01'  # len(brokers)
                    b'\x00\x00\x00\x00'  # node id # noqa
                    b'\x00\x09'  # len(host)
                        b'localhost'  # host
                    b'\x00\x00#\x84'  # port
                    b'\xff\xff'  # len(rack)
                b'\x00\x01'  # len(cluster_id)
                    b'a'  # cluster_id
                b'\x00\x00\x00\x00'  # controller_id
                b'\x00\x00\x00\x01'  # len(topic metadata)
                    b'\x00\x00'  # error code
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00'  # is_internal
                    b'\x00\x00\x00\x02'  # len(partition metadata)
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x00'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x01'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replicas
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
            )
        )
        self.assertEqual(cluster.brokers[0].host, b'localhost')
        self.assertEqual(cluster.brokers[0].port, 9092)
        self.assertEqual(cluster.topics[b'test'].partitions[0].leader,
                         cluster.brokers[0].id)
        self.assertEqual(cluster.topics[b'test'].partitions[0].replicas,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.topics[b'test'].partitions[0].isr,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.brokers[0].rack, None)
        self.assertEqual(cluster.throttle_time_ms, 0)
        self.assertEqual(cluster.controller_id, 0)
        self.assertEqual(cluster.cluster_id, b"a")
        self.assertEqual(cluster.topics[b'test'].is_internal, False)


class TestMetadataAPIV5(unittest2.TestCase):
    maxDiff = None

    def test_request(self):
        req = protocol.MetadataRequestV5()
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                # header
                b'\x00\x00\x00\x16'  # len(buffer)
                b'\x00\x03'  # ApiKey
                b'\x00\x05'  # api version
                b'\x00\x00\x00\x00'  # correlation id
                b'\x00\x07'  # len(client id)
                    b'pykafka'  # client id  # noqa
                # end header

                b'\xff\xff\xff\xff'  # len(topics)
                b'\x01'  # allow_topic_autocreation
            )
        )

    def test_response(self):
        cluster = protocol.MetadataResponseV5(
            buffer(
                b'\x00\x00\x00\x00'  # throttle_time_ms
                b'\x00\x00\x00\x01'  # len(brokers)
                    b'\x00\x00\x00\x00'  # node id # noqa
                    b'\x00\x09'  # len(host)
                        b'localhost'  # host
                    b'\x00\x00#\x84'  # port
                    b'\xff\xff'  # len(rack)
                b'\x00\x01'  # len(cluster_id)
                    b'a'  # cluster_id
                b'\x00\x00\x00\x00'  # controller_id
                b'\x00\x00\x00\x01'  # len(topic metadata)
                    b'\x00\x00'  # error code
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                    b'\x00'  # is_internal
                    b'\x00\x00\x00\x02'  # len(partition metadata)
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x00'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replica
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
                            b'\x00\x00\x00\x01'  # len(offline_replicas)
                                b'\x00\x00\x00\x00'  # offline_replicas
                        b'\x00\x00'  # partition error code
                            b'\x00\x00\x00\x01'  # partition id
                            b'\x00\x00\x00\x00'  # leader
                            b'\x00\x00\x00\x01'  # len(replicas)
                                b'\x00\x00\x00\x00'  # replicas
                            b'\x00\x00\x00\x01'  # len(isr)
                                b'\x00\x00\x00\x00'  # isr
                            b'\x00\x00\x00\x01'  # len(offline_replicas)
                                b'\x00\x00\x00\x00'  # offline_replicas
            )
        )
        self.assertEqual(cluster.brokers[0].host, b'localhost')
        self.assertEqual(cluster.brokers[0].port, 9092)
        self.assertEqual(cluster.topics[b'test'].partitions[0].leader,
                         cluster.brokers[0].id)
        self.assertEqual(cluster.topics[b'test'].partitions[0].replicas,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.topics[b'test'].partitions[0].isr,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.topics[b'test'].partitions[0].offline_replicas,
                         [cluster.brokers[0].id])
        self.assertEqual(cluster.brokers[0].rack, None)
        self.assertEqual(cluster.throttle_time_ms, 0)
        self.assertEqual(cluster.controller_id, 0)
        self.assertEqual(cluster.cluster_id, b"a")
        self.assertEqual(cluster.topics[b'test'].is_internal, False)


class TestProduceAPI(unittest2.TestCase):
    maxDiff = None

    test_messages = [
        protocol.Message(b'this is a test message', partition_key=b'asdf'),
        protocol.Message(b'this is a test message', partition_key=b'asdf',
                         timestamp=1497302164, protocol_version=1),
        protocol.Message(b'this is also a test message', partition_key=b'test_key'),
        protocol.Message(b"this doesn't have a partition key"),
    ]

    def test_request(self):
        message = self.test_messages[0]
        req = protocol.ProduceRequest()
        req.add_message(message, b'test', 0)
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b"\x00\x00\x00a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07pykafka"  # header
                b'\x00\x01'  # required acks
                b'\x00\x00\'\x10'  # timeout
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len (partitions)
                        b'\x00\x00\x00\x00'  # partition
                        b'\x00\x00\x004'  # message set size
                            b'\xff\xff\xff\xff\xff\xff\xff\xff'  # offset
                            b'\x00\x00\x00('  # message size
                                b'\x0e\x8a\x19O'  # crc
                                b'\x00'  # magic byte
                                b'\x00'  # attributes
                                b'\x00\x00\x00\x04'  # len(key)
                                    b'asdf'  # key
                                b'\x00\x00\x00\x16'  # len(value)
                                    b"this is a test message"  # value
            )
        )

    def test_request_message_timestamp(self):
        message = self.test_messages[1]
        req = protocol.ProduceRequest()
        req.add_message(message, b'test', 0)
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00i\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\x01'  # required acks
                b"\x00\x00\'\x10"  # timeout
                b"\x00\x00\x00\x01"  # len(topics)
                    b"\x00\x04"  # len(topic name) # noqa
                        b"test"  # topic name
                    b"\x00\x00\x00\x01"  # len(partitions)
                        b"\x00\x00\x00\x00"  # partition
                        b"\x00\x00\x00<"  # message set size
                            b"\xff\xff\xff\xff\xff\xff\xff\xff"  # offset
                            b"\x00\x00\x000"  # message size
                                b"Z\x92\x80\t"  # crc
                                b"\x01"  # magic byte (protocol version)
                                b"\x00"  # attributes
                                b"\x00\x00\x00\x00Y?\x04\x94"  # timestamp
                                b"\x00\x00\x00\x04"  # len(key)
                                    b"asdf"  # key
                                b"\x00\x00\x00\x16"  # len(value)
                                    b"this is a test message"  # value
            )
        )

    def test_gzip_compression(self):
        req = protocol.ProduceRequest(compression_type=CompressionType.GZIP)
        [req.add_message(m, b'test_gzip', 0) for m in self.test_messages]
        msg = req.get_bytes()
        self.assertEqual(len(msg), 230)  # this isn't a good test

    def test_snappy_compression(self):
        req = protocol.ProduceRequest(compression_type=CompressionType.SNAPPY)
        [req.add_message(m, b'test_snappy', 0) for m in self.test_messages]
        msg = req.get_bytes()
        self.assertEqual(len(msg), 240)  # this isn't a good test

    def test_partition_error(self):
        # Response has a UnknownTopicOrPartition error for test/0
        response = protocol.ProduceResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                            b'\x00\x03'  # error code
                            b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
            )
        )
        self.assertEqual(response.topics[b'test'][0].err, 3)

    def test_response(self):
        response = protocol.ProduceResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                            b'\x00\x00'  # error code
                            b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
            )
        )
        self.assertEqual(
            response.topics,
            {b'test': {0: protocol.ProducePartitionResponse(0, 2)}}
        )


class _FetchAPITestBase(object):
    maxDiff = None

    def msg_to_dict(self, msg):
        """Helper to extract data from Message slots"""
        attr_names = protocol.Message.__slots__
        f = operator.attrgetter(*attr_names)
        return dict(zip(attr_names, f(msg)))


class TestFetchAPI(unittest2.TestCase, _FetchAPITestBase):
    def _get_expected(self):
        return [{
            'partition_key': b'asdf',
            'compression_type': 0,
            'value': b'this is a test message',
            'offset': 0,
            'partition_id': 0,
            'produce_attempt': 0,
            'delivery_report_q': None,
            'timestamp': 0,
            'protocol_version': 0,
            'partition': None
        }, {
            'partition_key': b'test_key',
            'compression_type': 0,
            'value': b'this is also a test message',
            'offset': 1,
            'partition_id': 0,
            'produce_attempt': 0,
            'delivery_report_q': None,
            'timestamp': 0,
            'protocol_version': 0,
            'partition': None
        }, {
            'partition_key': None,
            'compression_type': 0,
            'value': b"this doesn't have a partition key",
            'offset': 2,
            'partition_id': 0,
            'produce_attempt': 0,
            'delivery_report_q': None,
            'timestamp': 0,
            'protocol_version': 0,
            'partition': None
        }]

    def test_request(self):
        preq = protocol.PartitionFetchRequest(b'test', 0, 1)
        req = protocol.FetchRequest(partition_requests=[preq, ])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00;\x00\x01\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\xff\xff\xff\xff'  # replica id
                b'\x00\x00\x03\xe8'  # max wait time
                b'\x00\x00\x04\x00'  # min bytes
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                            b'\x00\x00\x00\x00\x00\x00\x00\x01'  # fetch offset
                            b'\x00\x10\x00\x00'  # max bytes
            )
        )

    def test_partition_error(self):
        # Response has a UnknownTopicOrPartition error for test/0
        response = protocol.FetchResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # len(topic)
                b'\x00\x00\x00\x01'  # len (partitions)
                    b'\x00\x00\x00\x00'  # partition id
                    b'\x00\x03'  # error code
                        b'\x00\x00\x00\x00\x00\x00\x00\x02'  # highwater mark offset
                        b'\x00\x00\x00B'  # message set size
                            b'\x00\x00\x00\x00\x00\x00\x00\x01'  # offset
                            b'\x00\x00\x006'  # message size
                                b'\xa3 ^B'  # crc
                                b'\x00'  # magic byte
                                b'\x00'  # attributes
                                b'\x00\x00\x00\x12'  # len(key)
                                    b'test_partition_key'  # key
                                b'\x00\x00\x00\x16'  # len(value)
                                    b'this is a test message'  # value
            )
        )
        self.assertEqual(response.topics[b'test'][0].err, 3)

    def test_response(self):
        resp = protocol.FetchResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # len(topic)
                b'\x00\x00\x00\x01'  # len (partitions)
                    b'\x00\x00\x00\x00'  # partition id
                    b'\x00\x00'  # error code
                        b'\x00\x00\x00\x00\x00\x00\x00\x02'  # highwater mark offset
                        b'\x00\x00\x00B'  # message set size
                            b'\x00\x00\x00\x00\x00\x00\x00\x01'  # offset
                            b'\x00\x00\x006'  # message size
                                b'\xa3 ^B'  # crc
                                b'\x00'  # magic byte
                                b'\x00'  # attributes
                                b'\x00\x00\x00\x12'  # len(key)
                                    b'test_partition_key'  # key
                                b'\x00\x00\x00\x16'  # len(value)
                                    b'this is a test message'  # value
            )
        )
        self.assertEqual(len(resp.topics[b'test'][0].messages), 1)
        self.assertEqual(resp.topics[b'test'][0].max_offset, 2)
        message = resp.topics[b'test'][0].messages[0]
        self.assertEqual(message.value, b'this is a test message')
        self.assertEqual(message.partition_key, b'test_partition_key')
        self.assertEqual(message.compression_type, 0)
        self.assertEqual(message.offset, 1)

    def test_gzip_decompression(self):
        msg = b''.join([
            b"\x00\x00\x00\x01\x00\ttest_gzip\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x05\x00\x00\x00\xad\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00"
            b"\x00\xa1S\x82\x9d\xff\x00\x01\xff\xff\xff\xff\x00\x00\x00\x93\x1f\x8b\x08\x00\x00"
            b"\x00\x00\x00\x00\x00c`\x80\x03\r\xbe.I\x7f0\x8b%\xb18%\rH\x8b\x95dd\x16+\x00Q\xa2"
            b"BIjq\x89Bnjqqbz*T=#\x10\x1b\xb2\xf3\xcb\xf4\x81y\x1c \x15\xf1\xd9\xa9\x95@\xb6"
            b"4\\_Nq>v\xcdL@\xac\x7f\xb5(\xd9\x98\x81\xe1?\x10\x00y\x8a`M)\xf9\xa9\xc5y\xea%\n"
            b"\x19\x89e\xa9@\x9d\x05\x89E%\x99%\x99\xf9y\n\x10\x93A\x80\x19\x88\xcd/\x9a<1\xc5"
            b"\xb0\x97h#X\xc8\xb0\x1d\x00Bj\t\\*\x01\x00\x00\x00\x00\x00\x00"
        ])
        response = protocol.FetchResponse(msg)
        expected_data = self._get_expected()
        for i in range(len(expected_data)):
            self.assertDictEqual(
                self.msg_to_dict(response.topics[b'test_gzip'][0].messages[i]),
                expected_data[i])

    def test_snappy_decompression(self):
        msg = b''.join([
            b'\x00\x00\x00\x01'  # len(topics)
            b'\x00\x0b'  # len(topic name)
                b'test_snappy'  # topic name  # noqa
            b'\x00\x00\x00\x01'  # len(partitions)
                b'\x00\x00\x00\x00'  # partition
                    b'\x00\x00'  # error code
                    b'\x00\x00\x00\x00\x00\x00\x00\x03'  # highwater mark offset
                    b'\x00\x00\x00\xb5'  # message set size
                        b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
                        b'\x00\x00\x00\xa9'  # message size
                            b'\xc1\xf2\xa3\xe1'  # crc
                            b'\x00'  # magic bytes
                            b'\x02'  # attributes
                            b'\xff\xff\xff\xff'  # len(key)
                            b'\x00\x00\x00\x9b'  # len(value)
                                b'\x82SNAPPY\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x87\xac\x01\x00\x00\x19\x01\x10(\x0e\x8a\x19O\x05\x0fx\x04asdf\x00\x00\x00\x16this is a test message\x05$(\x00\x00\x01\x00\x00\x001\x07\x0f\x1c\x8e\x05\x10\x00\x08\x01"\x1c_key\x00\x00\x00\x1b\x158\x08lsoV=\x00H\x02\x00\x00\x00/\xd5rc3\x00\x00\xff\xff\xff\xff\x00\x00\x00!\x055ldoesn\'t have a partition key'  # value
        ])
        response = protocol.FetchResponse(msg)
        expected_data = self._get_expected()
        for i in range(len(expected_data)):
            self.assertDictEqual(
                self.msg_to_dict(response.topics[b'test_snappy'][0].messages[i]),
                expected_data[i])


class TestFetchAPIV1(unittest2.TestCase, _FetchAPITestBase):
    RESPONSE_CLASS = protocol.FetchResponseV1

    def _get_expected(self):
        return [{
            'partition_key': b'asdf',
            'compression_type': 0,
            'value': b'this is a test message',
            'offset': 0,
            'partition_id': 0,
            'produce_attempt': 0,
            'delivery_report_q': None,
            'timestamp': 1497393304998,
            'protocol_version': 1,
            'partition': None
        }, {
            'partition_key': b'test_key',
            'compression_type': 0,
            'value': b'this is also a test message',
            'offset': 1,
            'partition_id': 0,
            'produce_attempt': 0,
            'delivery_report_q': None,
            'timestamp': 1497393305005,
            'protocol_version': 1,
            'partition': None
        }, {
            'partition_key': None,
            'compression_type': 0,
            'value': b"this doesn't have a partition key",
            'offset': 2,
            'partition_id': 0,
            'produce_attempt': 0,
            'delivery_report_q': None,
            'timestamp': 1497393305013,
            'protocol_version': 1,
            'partition': None
        }, {
            'partition_key': b"test_key",
            'compression_type': 0,
            'value': b"this has a partition key and a timestamp",
            'offset': 3,
            'partition_id': 0,
            'produce_attempt': 0,
            'delivery_report_q': None,
            'timestamp': 1497302164,
            'protocol_version': 1,
            'partition': None
        }, {
            'partition_key': None,
            'compression_type': 0,
            'value': b"this has a timestamp",
            'offset': 4,
            'partition_id': 0,
            'produce_attempt': 0,
            'delivery_report_q': None,
            'timestamp': 1497302164,
            'protocol_version': 1,
            'partition': None
        }]

    def test_partition_error(self):
        # Response has a UnknownTopicOrPartition error for test/0
        response = self.RESPONSE_CLASS(
            buffer(
                b'\x00\x00\x00\x01'  # throttle time
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # len(topic)
                b'\x00\x00\x00\x01'  # len (partitions)
                    b'\x00\x00\x00\x00'  # partition id
                    b'\x00\x03'  # error code
                        b'\x00\x00\x00\x00\x00\x00\x00\x02'  # highwater mark offset
                        b'\x00\x00\x00B'  # message set size
                            b'\x00\x00\x00\x00\x00\x00\x00\x01'  # offset
                            b'\x00\x00\x006'  # message size
                                b'\xa3 ^B'  # crc
                                b'\x00'  # magic byte
                                b'\x00'  # attributes
                                b'\x00\x00\x00\x12'  # len(key)
                                    b'test_partition_key'  # key
                                b'\x00\x00\x00\x16'  # len(value)
                                    b'this is a test message'  # value
            )
        )
        self.assertEqual(response.topics[b'test'][0].err, 3)

    def test_response(self):
        resp = self.RESPONSE_CLASS(
            buffer(
                b'\x00\x00\x00\x01'  # throttle time
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # len(topic)
                b'\x00\x00\x00\x01'  # len (partitions)
                    b'\x00\x00\x00\x00'  # partition id
                    b'\x00\x00'  # error code
                        b'\x00\x00\x00\x00\x00\x00\x00\x02'  # highwater mark offset
                        b'\x00\x00\x00B'  # message set size
                            b'\x00\x00\x00\x00\x00\x00\x00\x01'  # offset
                            b'\x00\x00\x006'  # message size
                                b'\xa3 ^B'  # crc
                                b'\x00'  # magic byte
                                b'\x00'  # attributes
                                b'\x00\x00\x00\x12'  # len(key)
                                    b'test_partition_key'  # key
                                b'\x00\x00\x00\x16'  # len(value)
                                    b'this is a test message'  # value
            )
        )
        self.assertEqual(resp.throttle_time, 1)
        self.assertEqual(len(resp.topics[b'test'][0].messages), 1)
        self.assertEqual(resp.topics[b'test'][0].max_offset, 2)
        message = resp.topics[b'test'][0].messages[0]
        self.assertEqual(message.value, b'this is a test message')
        self.assertEqual(message.partition_key, b'test_partition_key')
        self.assertEqual(message.compression_type, 0)
        self.assertEqual(message.offset, 1)

    def test_gzip_decompression(self):
        msg = b''.join([
b'\x00\x00\x00\x00\x00\x00\x00\x01\x00\x0btest_gzip_5\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x05\x00\x00\x02\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00Z\xb1Z\xf4\xc3\x01\x01\x00\x00\x01\\\xa3\x98\x95\xa6\xff\xff\xff\xff\x00\x00\x00D\x1f\x8b\x08\x00\x00\x00\x00\x00\x00\x00c`\x80\x03\x03\x97?.{\x19\x81\x0c\xc6\x98\xc53\xa6.\x032X\x12\x8bS\xd2\x80\xb4XIFf\xb1\x02\x10%*\x94\xa4\x16\x97(\xe4\xa6\x16\x17\'\xa6\xa7\x02\x00N\xddm\x92<\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00`\xcdX\t\xed\x01\x01\x00\x00\x01\\\xa3\x98\x95\xad\xff\xff\xff\xff\x00\x00\x00J\x1f\x8b\x08\x00\x00\x00\x00\x00\x00\x00c`\x80\x03\xcb%\xf2\x01\x99\x8c@\x06c\xcc\xe2\x19S\xd7\x02\x19\x1c%\xa9\xc5%\xf1\xd9\xa9\x95@\xb6tIFf\xb1\x02\x10%\xe6\x14\xe7+$*\x80\xa4\x14rS\x8b\x8b\x13\xd3S\x01\xfe<~BE\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00b\xba\xb0lN\x01\x01\x00\x00\x01\\\xa3\x98\x95\xb5\xff\xff\xff\xff\x00\x00\x00L\x1f\x8b\x08\x00\x00\x00\x00\x00\x00\x00c`\x80\x03s\xfb\xad2\x07\x19\x81\x0c\xc6\x98\xc53\xa6n\xfd\x0f\x04@\x8ebIFf\xb1BJ~jq\x9ez\x89BFbY\xaaB\xa2BAbQIfIf~\x9eBvj%\x00\xc8f?\xe3C\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x03\x00\x00\x00l\x9f\xd7)\xa0\x01\x01\x00\x00\x00\x00Y?\x04\x94\xff\xff\xff\xff\x00\x00\x00V\x1f\x8b\x08\x00\x00\x00\x00\x00\x00\x00c`\x80\x037\x86wK\x8f0\x82\x99\x91\xf6,S\x80\x14GIjqI|vj%\x90\xadQ\x92\x91Y\xac\x90\x91X\xac\x90\xa8P\x90XT\x92Y\x92\x99\x9f\xa7\x00\x94SH\xccK\x01\x8a\x95d\xe6\x02\x15\'\xe6\x16\x00\x00\xac\xc0O.R\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00S\xd8"\xff7\x01\x01\x00\x00\x00\x00Y?\x04\x94\xff\xff\xff\xff\x00\x00\x00=\x1f\x8b\x08\x00\x00\x00\x00\x00\x00\x00c`\x80\x03\xad\x0f\xae\x97^2\x82\x99\x91\xf6,S\xfe\x03\x01\x90)R\x92\x91Y\xac\x90\x91X\xac\x90\xa8P\x92\x99\x9bZ\\\x92\x98[\x00\x00\x83\x0f\xe5\xc16\x00\x00\x00\x00\x00\x00\x00' # noqa
            ])
        response = self.RESPONSE_CLASS(msg)
        expected_data = self._get_expected()
        for i in range(len(expected_data)):
            self.assertDictEqual(
                self.msg_to_dict(response.topics[b'test_gzip_5'][0].messages[i]),
                expected_data[i])

    def test_snappy_decompression(self):
        msg = b''.join([
b"\x00\x00\x00\x00\x00\x00\x00\x01\x00\x0btest_snappy\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x05\x00\x00\x02=\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00a\xaa\x92\x19\xe2\x01\x02\x00\x00\x01\\\xa3\xa5\xab/\xff\xff\xff\xff\x00\x00\x00K\x82SNAPPY\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x007<\x00\x00\x19\x01\xc00\x86\xf55\x85\x01\x00\x00\x00\x01\\\xa3\xa5\xab/\x00\x00\x00\x04asdf\x00\x00\x00\x16this is a test message\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00jI\xfc\xe7\xf6\x01\x02\x00\x00\x01\\\xa3\xa5\xab\xa1\xff\xff\xff\xff\x00\x00\x00T\x82SNAPPY\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00@E\x00\x00\x19\x01\xe49\xd4\r\xf8v\x01\x00\x00\x00\x01\\\xa3\xa5\xab\xa1\x00\x00\x00\x08test_key\x00\x00\x00\x1bthis is also a test message\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00h\x80\t0\x95\x01\x02\x00\x00\x01\\\xa3\xa5\xab\xe1\xff\xff\xff\xff\x00\x00\x00R\x82SNAPPY\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00>C\x00\x00\x19\x01\xdc72\x98C\xe8\x01\x00\x00\x00\x01\\\xa3\xa5\xab\xe1\xff\xff\xff\xff\x00\x00\x00!this doesn't have a partition key\x00\x00\x00\x00\x00\x00\x00\x03\x00\x00\x00uJ\xab\xd0\xdf\x01\x02\x00\x00\x00\x00Y?\x04\x94\xff\xff\xff\xff\x00\x00\x00_\x82SNAPPY\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00KR\x00\x00\x19\x01\x14F\x00\xee\xa5\xc4\x01\x05\x10\xecY?\x04\x94\x00\x00\x00\x08test_key\x00\x00\x00(this has a partition key and a timestamp\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00Y\xfb!\x15\xce\x01\x02\x00\x00\x00\x00Y?\x04\x94\xff\xff\xff\xff\x00\x00\x00C\x82SNAPPY\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00/6\x00\x00\x19\x01\x14*\xf0E\xd2\xe9\x01\x05\x10|Y?\x04\x94\xff\xff\xff\xff\x00\x00\x00\x14this has a timestamp\x00\x00\x00\x00" # noqa
        ])
        response = self.RESPONSE_CLASS(msg)
        expected_data = self._get_expected()
        for i in range(len(expected_data)):
            returned = self.msg_to_dict(response.topics[b'test_snappy'][0].messages[i])
            expected = expected_data[i]
            # ignore timestamps if they were auto-set
            if i <= 2:
                returned.pop("timestamp")
                expected.pop("timestamp")
            self.assertDictEqual(returned, expected)


class TestFetchAPIV2(TestFetchAPIV1):
    RESPONSE_CLASS = protocol.FetchResponseV2


class TestListOffsetAPI(unittest2.TestCase):
    maxDiff = None

    def test_request(self):
        preq = protocol.PartitionOffsetRequest(b'test', 0, -1, 1)
        req = protocol.ListOffsetRequest(partition_requests=[preq, ])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x003\x00\x02\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\xff\xff\xff\xff'  # replica id
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                        b'\xff\xff\xff\xff\xff\xff\xff\xff'  # time
                        b'\x00\x00\x00\x01'  # max number of offsets
            )
        )

    def test_partition_error(self):
        # Response has a UnknownTopicOrPartition error for test/0
        response = protocol.ListOffsetResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partitoin
                        b'\x00\x03'  # error code
                        b'\x00\x00\x00\x01'  # len(offsets)
                            b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
            )
        )
        self.assertEqual(response.topics[b'test'][0].err, 3)

    def test_response(self):
        resp = protocol.ListOffsetResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partitoin
                        b'\x00\x00'  # error code
                        b'\x00\x00\x00\x01'  # len(offsets)
                            b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
            )
        )
        self.assertEqual(resp.topics[b'test'][0].offset, [2])


class TestListOffsetAPIV1(unittest2.TestCase):
    maxDiff = None

    def test_request(self):
        preq = protocol.PartitionOffsetRequest(b'test', 0, -1, 1)
        req = protocol.ListOffsetRequestV1(partition_requests=[preq, ])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                # header
                b'\x00\x00\x00/'  # len(buffer)
                b'\x00\x02'  # ApiKey
                b'\x00\x01'  # api version
                b'\x00\x00\x00\x00'  # correlation id
                b'\x00\x07'  # len(client id)
                    b'pykafka'  # client id  # noqa
                # end header
                b'\xff\xff\xff\xff'  # replica id
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                        b'\xff\xff\xff\xff\xff\xff\xff\xff'  # time
            )
        )

    def test_partition_error(self):
        # Response has a UnknownTopicOrPartition error for test/0
        response = protocol.ListOffsetResponseV1(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partitoin
                        b'\x00\x03'  # error code
                        b'\x00\x00\x00\x00\x00\x00\x00\x02'  # timestamp
                        b'\x00\x00\x00\x01'  # len(offsets)
                            b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
            )
        )
        self.assertEqual(response.topics[b'test'][0].err, 3)

    def test_response(self):
        resp = protocol.ListOffsetResponseV1(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name) # noqa
                        b'test'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partitoin
                        b'\x00\x00'  # error code
                        b'\x00\x00\x00\x00\x00\x00\x00\x02'  # timestamp
                        b'\x00\x00\x00\x01'  # len(offsets)
                            b'\x00\x00\x00\x00\x00\x00\x00\x02'  # offset
            )
        )
        self.assertEqual(resp.topics[b'test'][0].offset, [2])
        self.assertEqual(resp.topics[b'test'][0].timestamp, 2)


class TestOffsetCommitFetchAPI(unittest2.TestCase):
    maxDiff = None

    def test_consumer_metadata_request(self):
        req = protocol.GroupCoordinatorRequest(b'test')
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00\x17\x00\n\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\x04'  # len(group id)
                    b'test'  # group id # noqa
            )
        )

    def test_consumer_metadata_response(self):
        response = protocol.GroupCoordinatorResponse(
            buffer(
                b'\x00\x00'  # error code
                b'\x00\x00\x00\x00'  # coordinator id
                b'\x00\r'  # len(coordinator host)
                    b'emmett-debian'  # coordinator host # noqa
                b'\x00\x00#\x84'  # coordinator port
            )
        )
        self.assertEqual(response.coordinator_id, 0)
        self.assertEqual(response.coordinator_host, b'emmett-debian')
        self.assertEqual(response.coordinator_port, 9092)

    def test_offset_commit_request(self):
        preq = protocol.PartitionOffsetCommitRequest(
            b'test', 0, 68, 1426632066, b'testmetadata')
        req = protocol.OffsetCommitRequest(
            b'test', 1, b'pykafka', partition_requests=[preq, ])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00T\x00\x08\x00\x01\x00\x00\x00\x00\x00\x07pykafka'
                b'\x00\x04'  # len(consumer group id)
                    b'test'  # consumer group id # noqa
                b'\x00\x00\x00\x01'  # consumer group generation id
                b'\x00\x07'  # len(consumer id)
                    b'pykafka'  # consumer id
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x04'  # len(topic name)
                        b'test'  # topic name
                b'\x00\x00\x00\x01'  # len(partitions)
                    b'\x00\x00\x00\x00'  # partition
                        b'\x00\x00\x00\x00\x00\x00\x00D'  # offset
                        b'\x00\x00\x00\x00U\x08\xad\x82'  # timestamp
                        b'\x00\x0c'  # len(metadata)
                            b'testmetadata'  # metadata
            )
        )

    def test_offset_commit_response(self):
        response = protocol.OffsetCommitResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x0c'  # len(topic name) # noqa
                        b'emmett.dummy'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                            b'\x00\x00'  # error code
            )
        )
        self.assertEqual(response.topics[b'emmett.dummy'][0].err, 0)

    def test_offset_fetch_request(self):
        preq = protocol.PartitionOffsetFetchRequest(b'testtopic', 0)
        req = protocol.OffsetFetchRequest(b'test', partition_requests=[preq, ])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                # header
                b'\x00\x00\x00\x2e'  # len(buffer)
                b'\x00\x09'  # ApiKey
                b'\x00\x00'  # api version
                b'\x00\x00\x00\x00'  # correlation id
                b'\x00\x07'  # len(client id)
                    b'pykafka'  # client id
                # end header

                b'\x00\x04'  # len(consumer group)
                    b'test'  # consumer group # noqa
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x09'  # len(topic name)
                        b'testtopic'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
            )
        )

    def test_offset_fetch_response(self):
        response = protocol.OffsetFetchResponse(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x0c'  # len(topic name) # noqa
                        b'emmett.dummy'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                            b'\x00\x00\x00\x00\x00\x00\x00\x01'  # offset
                            b'\x00\x00'  # len(metadata)
                            b'\x00\x00'  # error code
            )
        )
        self.assertEqual(response.topics[b'emmett.dummy'][0].metadata, b'')
        self.assertEqual(response.topics[b'emmett.dummy'][0].offset, 1)


class TestOffsetCommitFetchAPIV2(unittest2.TestCase):
    maxDiff = None

    def test_offset_fetch_request(self):
        req = protocol.OffsetFetchRequestV2(b'test', partition_requests=[])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                # header
                b'\x00\x00\x00\x1b'  # len(buffer)
                b'\x00\x09'  # ApiKey
                b'\x00\x02'  # api version
                b'\x00\x00\x00\x00'  # correlation id
                b'\x00\x07'  # len(client id)
                    b'pykafka'  # client id
                # end header

                b'\x00\x04'  # len(consumer group)
                    b'test'  # consumer group # noqa
                b'\xff\xff\xff\xff'  # len(topics) = -1 for empty topics list
            )
        )

    def test_offset_fetch_response(self):
        response = protocol.OffsetFetchResponseV2(
            buffer(
                b'\x00\x00\x00\x01'  # len(topics)
                    b'\x00\x0c'  # len(topic name) # noqa
                        b'emmett.dummy'  # topic name
                    b'\x00\x00\x00\x01'  # len(partitions)
                        b'\x00\x00\x00\x00'  # partition
                            b'\x00\x00\x00\x00\x00\x00\x00\x01'  # offset
                            b'\x00\x00'  # len(metadata)
                            b'\x00\x00'  # error code
                b'\x00\x00'  # response level error_code
            )
        )
        self.assertEqual(response.topics[b'emmett.dummy'][0].metadata, b'')
        self.assertEqual(response.topics[b'emmett.dummy'][0].offset, 1)
        self.assertEqual(response.err, 0)


class TestGroupMembershipAPI(unittest2.TestCase):
    maxDiff = None

    def test_consumer_group_protocol_metadata(self):
        meta = protocol.ConsumerGroupProtocolMetadata()
        msg = meta.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00\x00'  # version
                b'\x00\x01'  # len(subscription)
                    b'\x00\n'  # len(topic name) # noqa
                        b'dummytopic'  # topic name
                    b'\x00\x00\x00\x0c'  # len(userdata)
                        b'testuserdata')  # userdata
        )

    def test_join_group_request(self):
        topic_name = b'abcdefghij'
        membership_protocol = RangeProtocol
        membership_protocol.metadata.topic_names = [topic_name]
        req = protocol.JoinGroupRequest(b'dummygroup', b'testmember', topic_name,
                                        membership_protocol)
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00h\x00\x0b\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\n'  # len(groupid)
                    b'dummygroup'  # groupid # noqa
                b'\x00\x00u0'  # session timeout
                b'\x00\n'  # len(memberid)
                    b'testmember'  # memberid
                b'\x00\x08'  # len(protocol type)
                    b'consumer'  # protocol type
                b'\x00\x00\x00\x01'  # len(group protocols)
                    b'\x00\x05'  # len(protocol name)
                        b'range'  # protocol name
                    b'\x00\x00\x00"'  # len(protocol metadata)
                        b'\x00\x00\x00\x00\x00\x01\x00\nabcdefghij\x00\x00\x00\x0ctestuserdata'  # protocol metadata
            )
        )

    def test_join_group_response(self):
        response = protocol.JoinGroupResponse(
            bytearray(
                b'\x00\x00'  # error code
                b'\x00\x00\x00\x01'  # generation id
                b'\x00\x17'  # len (group protocol)
                    b'dummyassignmentstrategy'  # group protocol # noqa
                b'\x00,'  # len(leader id)
                    b'pykafka-b2361322-674c-4e26-9194-305962636e57'  # leader id
                b'\x00,'  # len(member id)
                    b'pykafka-b2361322-674c-4e26-9194-305962636e57'  # member id
                b'\x00\x00\x00\x01'  # leb(members)
                    b'\x00,'  # len(member id)
                        b'pykafka-b2361322-674c-4e26-9194-305962636e57'  # member id
                    b'\x00\x00\x00"'  # len(member metadata)
                        b'\x00\x00\x00\x00\x00\x01\x00\ndummytopic\x00\x00\x00\x0ctestuserdata\x00\x00\x00\x00'  # member metadata
            )
        )
        self.assertEqual(response.generation_id, 1)
        self.assertEqual(response.group_protocol, b'dummyassignmentstrategy')
        member_id = b'pykafka-b2361322-674c-4e26-9194-305962636e57'
        self.assertEqual(response.leader_id, member_id)
        self.assertEqual(response.member_id, member_id)
        self.assertTrue(member_id in response.members)
        metadata = response.members[member_id]
        self.assertEqual(metadata.version, 0)
        self.assertEqual(metadata.topic_names, [b"dummytopic"])
        self.assertEqual(metadata.user_data, b"testuserdata")

    def test_member_assignment_construction(self):
        assignment = protocol.MemberAssignment([(b"mytopic1", [3, 5, 7, 9]),
                                                (b"mytopic2", [2, 4, 6, 8])])
        msg = assignment.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x01'  # version
                b'\x00\x00\x00\x02'  # len(partition assignment)
                    b'\x00\x08'  # len(topic) # noqa
                        b'mytopic1'  # topic
                    b'\x00\x00\x00\x04'  # len(partitions)
                        b'\x00\x00\x00\x03'  # partition
                        b'\x00\x00\x00\x05'  # partition
                        b'\x00\x00\x00\x07'  # partition
                        b'\x00\x00\x00\t'  # partition
                    b'\x00\x08'  # len(topic)
                        b'mytopic2'  # topic
                    b'\x00\x00\x00\x04'  # len(partitions)
                        b'\x00\x00\x00\x02'  # partition
                        b'\x00\x00\x00\x04'  # partition
                        b'\x00\x00\x00\x06'  # partition
                        b'\x00\x00\x00\x08'  # partition
            )
        )

    def test_sync_group_request(self):
        req = protocol.SyncGroupRequest(
            b'dummygroup', 1, b'testmember1',
            [
                (b"a", protocol.MemberAssignment([(b"mytopic1", [3, 5, 7, 9]),
                                                  (b"mytopic2", [3, 5, 7, 9])])),
                (b"b", protocol.MemberAssignment([(b"mytopic1", [2, 4, 6, 8]),
                                                  (b"mytopic2", [2, 4, 6, 8])]))
            ])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00\xc4\x00\x0e\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\n'  # len(group id)
                    b'dummygroup'  # group id # noqa
                b'\x00\x00\x00\x01'  # generation id
                b'\x00\x0b'  # len(member id)
                    b'testmember1'  # member id
                b'\x00\x00\x00\x02'  # len(group assignment)
                    b'\x00\x01'  # len(member id)
                        b'a'  # member id
                    b'\x00\x00\x00B'  # len(member assignment)
                        b'\x00\x01\x00\x00\x00\x02\x00\x08mytopic1\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x05\x00\x00\x00\x07\x00\x00\x00\t\x00\x08mytopic2\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x05\x00\x00\x00\x07\x00\x00\x00\t'  # member assignment
                    b'\x00\x01'  # len(member id)
                        b'b'  # member id
                    b'\x00\x00\x00B'  # len(member assignment)
                        b'\x00\x01\x00\x00\x00\x02\x00\x08mytopic1\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x04\x00\x00\x00\x06\x00\x00\x00\x08\x00\x08mytopic2\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x04\x00\x00\x00\x06\x00\x00\x00\x08'  # member assignment
            )
        )

    def test_sync_group_response(self):
        response = protocol.SyncGroupResponse(
            bytearray(
                b'\x00\x00'  # error code
                b'\x00\x00\x00H'  # len(member assignment)
                    b'\x00\x01\x00\x00\x00\x01\x00\x14testtopic_replicated\x00\x00\x00\n\x00\x00\x00\x06\x00\x00\x00\x07\x00\x00\x00\x08\x00\x00\x00\t\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x05,pyk'  # member assignment # noqa
            )
        )
        self.assertEqual(response.error_code, 0)
        expected_assignment = [(b'testtopic_replicated', [6, 7, 8, 9, 0, 1, 2, 3, 4, 5])]
        self.assertEqual(response.member_assignment.partition_assignment,
                         expected_assignment)

    def test_heartbeat_request(self):
        req = protocol.HeartbeatRequest(b'dummygroup', 1, b'testmember')
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00-\x00\x0c\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\n'  # len(group id)
                    b'dummygroup'  # group id # noqa
                b'\x00\x00\x00\x01'  # generation id
                b'\x00\n'  # len(member id)
                    b'testmember'  # member id
            )
        )

    def test_heartbeat_response(self):
        response = protocol.HeartbeatResponse(
            bytearray(
                b'\x00\x00'  # error code
            )
        )
        self.assertEqual(response.error_code, 0)

    def test_leave_group_request(self):
        req = protocol.LeaveGroupRequest(b'dummygroup', b'testmember')
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00)\x00\r\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\n'  # len(group id)
                    b'dummygroup'  # group id # noqa
                b'\x00\n'  # len(member id)
                    b'testmember'  # member id
            )
        )

    def test_leave_group_response(self):
        response = protocol.LeaveGroupResponse(
            bytearray(
                b'\x00\x00'  # error code
            )
        )
        self.assertEqual(response.error_code, 0)


class TestAdministrativeAPI(unittest2.TestCase):
    maxDiff = None

    def test_list_groups_request(self):
        req = protocol.ListGroupsRequest()
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00\x11\x00\x10\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
            )
        )

    def test_list_groups_response(self):
        response = protocol.ListGroupsResponse(
            bytearray(
                b'\x00\x00'  # error code
                b'\x00\x00\x00\x01'  # len(groups)
                    b'\x00\t'  # len(group_id) # noqa
                        b'testgroup'  # group_id
                    b'\x00\x08'  # len(protocol_type)
                        b'consumer'  # protocol_type
            )
        )
        self.assertEqual(len(response.groups), 1)
        group = response.groups[b"testgroup"]
        self.assertEqual(group.group_id, b"testgroup")
        self.assertEqual(group.protocol_type, b"consumer")

    def test_describe_groups_request(self):
        req = protocol.DescribeGroupsRequest([b'testgroup'])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00 \x00\x0f\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\x00\x00\x01'  # len(group_ids)
                    b'\x00\t'  # len(group_id) # noqa
                        b'testgroup'  # group_id
            )
        )

    def test_describe_groups_response(self):
        response = protocol.DescribeGroupsResponse(
            bytearray(
                b'\x00\x00\x00\x01'  # len(groups)
                    b'\x00\x00'  # error code # noqa
                    b'\x00\t'  # len(group_id)
                        b'testgroup'  # group_id
                    b'\x00\x06'  # len(state)
                        b'Stable'  # state
                    b'\x00\x08'  # len(protocol_type)
                        b'consumer'  # protocol_type
                    b'\x00\x05'  # len(protocol)
                        b'range'  # protocol
                    b'\x00\x00\x00\x01'  # len(members)
                        b'\x00,'  # len(member_id)
                            b'pykafka-d42426fb-c295-4cd9-b585-6dd79daf3afe'  # member_id
                        b'\x00\x07'  # len(client_id)
                            b'pykafka'  # client_id
                        b'\x00\n'  # len(client_host)
                            b'/127.0.0.1'  # client_host
                        b'\x00\x00\x00"'  # len(member_metadata)
                            b'\x00\x00\x00\x00\x00\x01\x00\ndummytopic\x00\x00\x00\x0ctestuserdata'
                        b'\x00\x00\x00H'  # len(member_assignment)
                            b'\x00\x01\x00\x00\x00\x01\x00\x14testtopic_replicated'  # member_assignment
                            b'\x00\x00\x00\n\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00'
                            b'\x02\x00\x00\x00\x08\x00\x00\x00\x03\x00\x00\x00\t'
                            b'\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00\x06'
                            b'\x00\x00\x00\x07\x00\x00\x00\x00'
            )
        )
        self.assertTrue(b'testgroup' in response.groups)
        group_response = response.groups[b'testgroup']
        self.assertEqual(group_response.error_code, 0)
        self.assertEqual(group_response.group_id, b'testgroup')
        self.assertEqual(group_response.state, b'Stable')
        self.assertEqual(group_response.protocol_type, b'consumer')
        self.assertEqual(group_response.protocol, b'range')
        member_id = b'pykafka-d42426fb-c295-4cd9-b585-6dd79daf3afe'
        self.assertTrue(member_id in group_response.members)
        member = group_response.members[member_id]
        self.assertEqual(member.member_id, member_id)
        self.assertEqual(member.client_id, b'pykafka')
        self.assertEqual(member.client_host, b'/127.0.0.1')
        metadata = member.member_metadata
        self.assertEqual(metadata.version, 0)
        self.assertEqual(metadata.topic_names, [b"dummytopic"])
        self.assertEqual(metadata.user_data, b"testuserdata")
        assignment = member.member_assignment
        self.assertEqual(assignment.version, 1)
        self.assertEqual(assignment.partition_assignment,
                         [(b'testtopic_replicated', [0, 1, 2, 8, 3, 9, 4, 5, 6, 7])])

    def test_create_topics_request(self):
        req = protocol.CreateTopicsRequest([protocol.CreateTopicRequest(b'mycooltopic', 4,
                                                                        2, [], [])])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x004\x00\x13\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\x00\x00\x01'  # len(topic_reqs)
                    b'\x00\x0b'  # len(topic)  # noqa
                        b'mycooltopic'  # topic
                    b'\x00\x00\x00\x04'  # num_partitions
                    b'\x00\x02'  # replication_factor
                    b'\x00\x00\x00\x00'  # len(replica_assignment)
                    b'\x00\x00\x00\x00'  # len(config_entries)
                b'\x00\x00\x00\x00'  # timeout
            )
        )

    def test_create_topics_response(self):
        response = protocol.CreateTopicsResponse(  # noqa
            bytearray(
                b'\x00\x00\x00\x01'  # len(topic_errors)
                    b'\x00\t'  # len(topic) # noqa
                        b'testtopic'  # topic_name
                    b'\x00\x00'  # error_code
            )
        )

    def test_delete_topics_request(self):
        req = protocol.DeleteTopicsRequest([b'mycooltopic'])
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00&\x00\x14\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
                b'\x00\x00\x00\x01'  # len(topic_error_codes)  # noqa
                    b'\x00\x0b'  # len(topic)
                        b'mycooltopic'  # topic
                b'\x00\x00\x00\x00'  # timeout
            )
        )

    def test_delete_topics_response(self):
        response = protocol.DeleteTopicsResponse(  # noqa
            bytearray(
                b'\x00\x00\x00\x01'  # len(topic_errors)
                    b'\x00\t'  # len(topic) # noqa
                        b'testtopic'  # topic_name
                    b'\x00\x00'  # error_code
            )
        )

    def test_api_versions_request(self):
        req = protocol.ApiVersionsRequest()
        msg = req.get_bytes()
        self.assertEqual(
            msg,
            bytearray(
                b'\x00\x00\x00\x11\x00\x12\x00\x00\x00\x00\x00\x00\x00\x07pykafka'  # header
            )
        )

    def test_api_versions_response(self):
        response = protocol.ApiVersionsResponse(
            bytearray(
                b'\x00\x00'  # error_code
                b'\x00\x00\x00\x15'  # len(api_versions)
                    b'\x00\x00\x00\x00\x00\x02'  # api_key, min_version, max_version # noqa
                    b'\x00\x01\x00\x00\x00\x03'
                    b'\x00\x02\x00\x00\x00\x01'
                    b'\x00\x03\x00\x00\x00\x02'
                    b'\x00\x04\x00\x00\x00\x00'
                    b'\x00\x05\x00\x00\x00\x00'
                    b'\x00\x06\x00\x00\x00\x02'
                    b'\x00\x07\x00\x01\x00\x01'
                    b'\x00\x08\x00\x00\x00\x02'
                    b'\x00\t\x00\x00\x00\x01'
                    b'\x00\n\x00\x00\x00\x00'
                    b'\x00\x0b\x00\x00\x00\x01'
                    b'\x00\x0c\x00\x00\x00\x00'
                    b'\x00\r\x00\x00\x00\x00'
                    b'\x00\x0e\x00\x00\x00\x00'
                    b'\x00\x0f\x00\x00\x00\x00'
                    b'\x00\x10\x00\x00\x00\x00'
                    b'\x00\x11\x00\x00\x00\x00'
                    b'\x00\x12\x00\x00\x00\x00'
                    b'\x00\x13\x00\x00\x00\x00'
                    b'\x00\x14\x00\x00\x00\x00'
                    b'\x00\x00\x00\x00'  # ???
            )
        )
        self.assertEqual(len(response.api_versions), 21)
        self.assertEqual(response.api_versions[1].max, 3)
        self.assertEqual(response.api_versions[5].min, 0)
        self.assertEqual(response.api_versions[12].key, 12)


if __name__ == '__main__':
    unittest2.main()



"""
Author: Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["handle_partition_responses", "raise_error"]
from collections import defaultdict
from .compat import iteritems


def handle_partition_responses(error_handlers,
                               parts_by_error=None,
                               success_handler=None,
                               response=None,
                               partitions_by_id=None):
    """Call the appropriate handler for each errored partition

    :param error_handlers: mapping of error code to handler
    :type error_handlers: dict {int: callable(parts)}
    :param parts_by_error: a dict of partitions grouped by error code
    :type parts_by_error: dict
        {int: iterable(:class:`pykafka.simpleconsumer.OwnedPartition`)}
    :param success_handler: function to call for successful partitions
    :type success_handler: callable accepting an iterable of partition responses
    :param response: a Response object containing partition responses
    :type response: :class:`pykafka.protocol.Response`
    :param partitions_by_id: a dict mapping partition ids to OwnedPartition
        instances
    :type partitions_by_id: dict
        {int: :class:`pykafka.simpleconsumer.OwnedPartition`}
    """
    if parts_by_error is None:
        parts_by_error = build_parts_by_error(response, partitions_by_id)

    for errcode, parts in iteritems(parts_by_error):
        if errcode != 0:
            error_handlers[errcode](parts)
        elif success_handler is not None:
            success_handler(parts)

    return parts_by_error


def build_parts_by_error(response, partitions_by_id):
    """Separate the partitions from a response by their error code

    :param response: a Response object containing partition responses
    :type response: :class:`pykafka.protocol.Response`
    :param partitions_by_id: a dict mapping partition ids to OwnedPartition
        instances
    :type partitions_by_id: dict
        {int: :class:`pykafka.simpleconsumer.OwnedPartition`}
    """
    # group partition responses by error code
    parts_by_error = defaultdict(list)
    for topic_name in response.topics.keys():
        for partition_id, pres in iteritems(response.topics[topic_name]):
            if partitions_by_id is not None and partition_id in partitions_by_id:
                owned_partition = partitions_by_id[partition_id]
                parts_by_error[pres.err].append((owned_partition, pres))
    return parts_by_error


def raise_error(error, info=""):
    """Raise the given error"""
    raise error(info)


def valid_int(param, allow_zero=False, allow_negative=False):
    """Validate that param is an integer, raise an exception if not"""
    pt = param
    try:  # a very permissive integer typecheck
        pt += 1
    except TypeError:
        raise TypeError(
            "Expected integer but found argument of type '{}'".format(type(param)))
    if not allow_negative and param < 0:
        raise ValueError("Expected nonnegative number but got '{}'".format(param))
    if not allow_zero and param == 0:
        raise ValueError("Expected nonzero number but got '{}'".format(param))
    return param



from __future__ import division
"""
Author: Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["SimpleConsumer"]
import datetime as dt
import itertools
import logging
import json
import socket
import sys
import threading
import time
from collections import defaultdict
import weakref

from six import reraise

from .common import OffsetType, EPOCH
from .utils.compat import (Queue, Empty, iteritems, itervalues,
                           range, iterkeys, get_bytes, get_string)
from .exceptions import (UnknownError, OffsetOutOfRangeError, UnknownTopicOrPartition,
                         OffsetMetadataTooLarge, GroupLoadInProgress,
                         NotCoordinatorForGroup, SocketDisconnectedError,
                         ConsumerStoppedException, KafkaException,
                         NotLeaderForPartition, OffsetRequestFailedError,
                         RequestTimedOut, UnknownMemberId, RebalanceInProgress,
                         IllegalGeneration, ERROR_CODES, UnicodeException)
from .protocol import (PartitionFetchRequest, PartitionOffsetCommitRequest,
                       PartitionOffsetFetchRequest, PartitionOffsetRequest)
from .utils.error_handlers import (handle_partition_responses, raise_error,
                                   build_parts_by_error, valid_int)


log = logging.getLogger(__name__)
MAGIC_OFFSETS = [OffsetType.EARLIEST, OffsetType.LATEST]


class SimpleConsumer(object):
    """
    A non-balancing consumer for Kafka
    """
    def __init__(self,
                 topic,
                 cluster,
                 consumer_group=None,
                 partitions=None,
                 fetch_message_max_bytes=1024 * 1024,
                 num_consumer_fetchers=1,
                 auto_commit_enable=False,
                 auto_commit_interval_ms=60 * 1000,
                 queued_max_messages=2000,
                 fetch_min_bytes=1,
                 fetch_error_backoff_ms=500,
                 fetch_wait_max_ms=100,
                 offsets_channel_backoff_ms=1000,
                 offsets_commit_max_retries=5,
                 auto_offset_reset=OffsetType.EARLIEST,
                 consumer_timeout_ms=-1,
                 auto_start=True,
                 reset_offset_on_start=False,
                 compacted_topic=False,
                 generation_id=-1,
                 consumer_id=b'',
                 deserializer=None,
                 reset_offset_on_fetch=True):
        """Create a SimpleConsumer.

        Settings and default values are taken from the Scala
        consumer implementation.  Consumer group is included
        because it's necessary for offset management, but doesn't imply
        that this is a balancing consumer. Use a BalancedConsumer for
        that.

        :param topic: The topic this consumer should consume
        :type topic: :class:`pykafka.topic.Topic`
        :param cluster: The cluster to which this consumer should connect
        :type cluster: :class:`pykafka.cluster.Cluster`
        :param consumer_group: The name of the consumer group this consumer
            should use for offset committing and fetching.
        :type consumer_group: str
        :param partitions: Existing partitions to which to connect
        :type partitions: Iterable of :class:`pykafka.partition.Partition`
        :param fetch_message_max_bytes: The number of bytes of messages to
            attempt to fetch
        :type fetch_message_max_bytes: int
        :param num_consumer_fetchers: The number of workers used to make
            FetchRequests
        :type num_consumer_fetchers: int
        :param auto_commit_enable: If true, periodically commit to kafka the
            offset of messages already returned from consume() calls. Requires that
            `consumer_group` is not `None`.
        :type auto_commit_enable: bool
        :param auto_commit_interval_ms: The frequency (in milliseconds) at which the
            consumer offsets are committed to kafka. This setting is ignored if
            `auto_commit_enable` is `False`.
        :type auto_commit_interval_ms: int
        :param queued_max_messages: Maximum number of messages buffered for
            consumption per partition
        :type queued_max_messages: int
        :param fetch_min_bytes: The minimum amount of data (in bytes) the server
            should return for a fetch request. If insufficient data is available
            the request will block until sufficient data is available.
        :type fetch_min_bytes: int
        :param fetch_error_backoff_ms: The amount of time (in milliseconds) that
            the consumer should wait before retrying after an error. Errors include
            absence of data (`RD_KAFKA_RESP_ERR__PARTITION_EOF`), so this can slow
            a normal fetch scenario. Only used by the native consumer
            (`RdKafkaSimpleConsumer`).
        :type fetch_error_backoff_ms: int
        :type fetch_error_backoff_ms: int
        :param fetch_wait_max_ms: The maximum amount of time (in milliseconds)
            the server will block before answering the fetch request if there
            isn't sufficient data to immediately satisfy `fetch_min_bytes`.
        :type fetch_wait_max_ms: int
        :param offsets_channel_backoff_ms: Backoff time (in milliseconds) to
            retry offset commits/fetches
        :type offsets_channel_backoff_ms: int
        :param offsets_commit_max_retries: Retry the offset commit up to this
            many times on failure.
        :type offsets_commit_max_retries: int
        :param auto_offset_reset: What to do if an offset is out of range. This
            setting indicates how to reset the consumer's internal offset
            counter when an `OffsetOutOfRangeError` is encountered.
        :type auto_offset_reset: :class:`pykafka.common.OffsetType`
        :param consumer_timeout_ms: Amount of time (in milliseconds) the
            consumer may spend without messages available for consumption
            before returning None.
        :type consumer_timeout_ms: int
        :param auto_start: Whether the consumer should begin communicating
            with kafka after __init__ is complete. If false, communication
            can be started with `start()`.
        :type auto_start: bool
        :param reset_offset_on_start: Whether the consumer should reset its
            internal offset counter to `self._auto_offset_reset` and commit that
            offset immediately upon starting up
        :type reset_offset_on_start: bool
        :param compacted_topic: Set to read from a compacted topic. Forces
            consumer to use less stringent message ordering logic because compacted
            topics do not provide offsets in strict incrementing order.
        :type compacted_topic: bool
        :param generation_id: Deprecated::2.7 Do not set if directly instantiating
            SimpleConsumer. The generation id with which to make group requests
        :type generation_id: int
        :param consumer_id: Deprecated::2.7 Do not set if directly instantiating
            SimpleConsumer. The identifying string to use for this consumer on
            group requests
        :type consumer_id: bytes
        :param deserializer: A function defining how to deserialize messages returned
            from Kafka. A function with the signature d(value, partition_key) that
            returns a tuple of (deserialized_value, deserialized_partition_key). The
            arguments passed to this function are the bytes representations of a
            message's value and partition key, and the returned data should be these
            fields transformed according to the client code's serialization logic.
            See `pykafka.utils.__init__` for stock implemtations.
        :type deserializer: function
        :param reset_offset_on_fetch: Whether to update offsets during fetch_offsets.
               Disable for read-only use cases to prevent side-effects.
        :type reset_offset_on_fetch: bool
        """
        self._running = False
        self._cluster = cluster
        self._consumer_group = None
        if consumer_group:
            try:
                self._consumer_group = get_string(consumer_group).encode('ascii')
            except UnicodeEncodeError:
                raise UnicodeException("Consumer group name '{}' contains non-ascii "
                                       "characters".format(consumer_group))
        self._topic = topic
        self._fetch_message_max_bytes = valid_int(fetch_message_max_bytes)
        self._fetch_min_bytes = valid_int(fetch_min_bytes)
        self._queued_max_messages = valid_int(queued_max_messages)
        self._num_consumer_fetchers = valid_int(num_consumer_fetchers)
        self._fetch_wait_max_ms = valid_int(fetch_wait_max_ms, allow_zero=True)
        self._consumer_timeout_ms = valid_int(consumer_timeout_ms,
                                              allow_zero=True, allow_negative=True)
        self._offsets_channel_backoff_ms = valid_int(offsets_channel_backoff_ms)
        self._auto_offset_reset = auto_offset_reset
        offsets_commit_max_retries = valid_int(offsets_commit_max_retries,
                                               allow_zero=True)
        self._offsets_commit_max_retries = offsets_commit_max_retries
        # not directly configurable
        self._offsets_fetch_max_retries = offsets_commit_max_retries
        self._offsets_reset_max_retries = offsets_commit_max_retries
        self._auto_start = auto_start
        self._reset_offset_on_start = reset_offset_on_start
        self._is_compacted_topic = compacted_topic
        self._generation_id = -1
        self._consumer_id = b''
        self._deserializer = deserializer
        self._reset_offset_on_fetch = reset_offset_on_fetch

        # incremented for any message arrival from any partition
        # the initial value is 0 (no messages waiting)
        self._messages_arrived = self._cluster.handler.Semaphore(value=0)
        self._slot_available = self._cluster.handler.Event()

        self._auto_commit_enable = auto_commit_enable
        self._auto_commit_interval_ms = valid_int(auto_commit_interval_ms)
        self._last_auto_commit = time.time()
        self._worker_exception = None
        self._update_lock = self._cluster.handler.Lock()

        self._discover_group_coordinator()

        if partitions is not None:
            self._partitions = {p: OwnedPartition(p,
                                                  self._cluster.handler,
                                                  self._messages_arrived,
                                                  self._is_compacted_topic)
                                for p in partitions}
        else:
            self._partitions = {topic.partitions[k]:
                                OwnedPartition(p,
                                               self._cluster.handler,
                                               self._messages_arrived,
                                               self._is_compacted_topic)
                                for k, p in iteritems(topic.partitions)}
        self._partitions_by_id = {p.partition.id: p
                                  for p in itervalues(self._partitions)}
        # Organize partitions by leader for efficient queries
        self._setup_partitions_by_leader()
        self.partition_cycle = itertools.cycle(self._partitions.values())

        self._default_error_handlers = self._build_default_error_handlers()

        if self._auto_start:
            self.start()

    @property
    def consumer_id(self):
        return self._consumer_id

    @consumer_id.setter
    def consumer_id(self, value):
        self._consumer_id = value
        for op in itervalues(self._partitions):
            op.set_consumer_id(self._consumer_id)

    @property
    def generation_id(self):
        return self._generation_id

    @generation_id.setter
    def generation_id(self, value):
        self._generation_id = valid_int(value, allow_zero=True,
                                        allow_negative=True)

    def __repr__(self):
        return "<{module}.{name} at {id_} (consumer_group={group})>".format(
            module=self.__class__.__module__,
            name=self.__class__.__name__,
            id_=hex(id(self)),
            group=self._consumer_group
        )

    def _raise_worker_exceptions(self):
        """Raises exceptions encountered on worker threads"""
        if self._worker_exception is not None:
            reraise(*self._worker_exception)

    def _update(self):
        """Update the consumer and cluster after an ERROR_CODE"""
        # only allow one thread to be updating the consumer at a time
        with self._update_lock:
            self._cluster.update()
            self._setup_partitions_by_leader()
            self._discover_group_coordinator()

    def start(self):
        """Begin communicating with Kafka, including setting up worker threads

        Fetches offsets, starts an offset autocommitter worker pool, and
        starts a message fetcher worker pool.
        """
        self._running = True

        # Figure out which offset wer're starting on
        if self._reset_offset_on_start:
            self.reset_offsets()
        elif self._consumer_group is not None:
            self.fetch_offsets()

        self._fetch_workers = self._setup_fetch_workers()
        if self._auto_commit_enable:
            self._autocommit_worker_thread = self._setup_autocommit_worker()

        self._raise_worker_exceptions()

    def _setup_partitions_by_leader(self):
        self._partitions_by_leader = defaultdict(list)
        for p in itervalues(self._partitions):
            self._partitions_by_leader[p.partition.leader].append(p)

    def _build_default_error_handlers(self):
        """Set up the error handlers to use for partition errors."""
        self = weakref.proxy(self)

        def _handle_OffsetOutOfRangeError(parts):
            log.info("Resetting offsets in response to OffsetOutOfRangeError")
            self.reset_offsets(
                partition_offsets=[(owned_partition.partition, self._auto_offset_reset)
                                   for owned_partition, pres in parts]
            )

        def _handle_RequestTimedOut(parts):
            log.info("Continuing in response to RequestTimedOut")

        def _handle_NotCoordinatorForGroup(parts):
            log.info("Updating cluster in response to NotCoordinatorForGroup")
            self._update()

        def _handle_NotLeaderForPartition(parts):
            log.info("Updating cluster in response to NotLeaderForPartition")
            self._update()

        def _handle_GroupLoadInProgress(parts):
            log.info("Continuing in response to GroupLoadInProgress")

        def _handle_IllegalGeneration(parts):
            log.info("Continuing in response to IllegalGeneration")

        def _handle_UnknownMemberId(parts):
            log.info("Continuing in response to UnknownMemberId")

        def _handle_UnknownError(parts):
            log.info("Continuing in response to UnknownError")

        def _handle_RebalanceInProgress(parts):
            log.info("Continuing in response to RebalanceInProgress")

        return {
            UnknownTopicOrPartition.ERROR_CODE: lambda p: raise_error(UnknownTopicOrPartition),
            UnknownError.ERROR_CODE: _handle_UnknownError,
            OffsetOutOfRangeError.ERROR_CODE: _handle_OffsetOutOfRangeError,
            NotLeaderForPartition.ERROR_CODE: _handle_NotLeaderForPartition,
            OffsetMetadataTooLarge.ERROR_CODE: lambda p: raise_error(OffsetMetadataTooLarge),
            NotCoordinatorForGroup.ERROR_CODE: _handle_NotCoordinatorForGroup,
            RequestTimedOut.ERROR_CODE: _handle_RequestTimedOut,
            GroupLoadInProgress.ERROR_CODE: _handle_GroupLoadInProgress,
            UnknownMemberId.ERROR_CODE: _handle_UnknownMemberId,
            RebalanceInProgress.ERROR_CODE: _handle_RebalanceInProgress,
            IllegalGeneration.ERROR_CODE: _handle_IllegalGeneration
        }

    def _discover_group_coordinator(self):
        """Set the group coordinator for this consumer.

        If a consumer group is not supplied to __init__, this method does nothing
        """
        if self._consumer_group is not None:
            self._group_coordinator = self._cluster.get_group_coordinator(self._consumer_group)

    @property
    def topic(self):
        """The topic this consumer consumes"""
        return self._topic

    @property
    def partitions(self):
        """A list of the partitions that this consumer consumes"""
        return {id_: partition.partition
                for id_, partition in iteritems(self._partitions_by_id)}

    @property
    def held_offsets(self):
        """Return a map from partition id to held offset for each partition"""
        return {p.partition.id:
                (OffsetType.EARLIEST if p.last_offset_consumed == -1
                 else p.last_offset_consumed)
                for p in itervalues(self._partitions_by_id)}

    def __del__(self):
        """Stop consumption and workers when object is deleted"""
        log.debug("Finalising {}".format(self))
        if self._running:
            self.stop()

    def cleanup(self):
        if not self._slot_available.is_set():
            self._slot_available.set()

    def stop(self):
        """Flag all running workers for deletion."""
        self._running = False
        if self._auto_commit_enable and self._consumer_group is not None:
            self.commit_offsets()
        # unblock a waiting consume() call
        if self._messages_arrived is not None:
            self._messages_arrived.release()

    def _setup_autocommit_worker(self):
        """Start the autocommitter thread"""
        self = weakref.proxy(self)

        def autocommitter():
            while True:
                try:
                    if not self._running:
                        break
                    if self._auto_commit_enable:
                        self._auto_commit()
                    self._cluster.handler.sleep(self._auto_commit_interval_ms / 1000)
                except ReferenceError:
                    break
                except Exception:
                    # surface all exceptions to the main thread
                    self._worker_exception = sys.exc_info()
                    break
            log.debug("Autocommitter thread exiting")
        log.debug("Starting autocommitter thread")
        return self._cluster.handler.spawn(autocommitter, name="pykafka.SimpleConsumer.autocommiter")

    def _setup_fetch_workers(self):
        """Start the fetcher threads"""
        # NB this gets overridden in rdkafka.RdKafkaSimpleConsumer
        self = weakref.proxy(self)

        def fetcher():
            while True:
                try:
                    if not self._running:
                        break
                    self.fetch()
                    self._cluster.handler.sleep(.01)
                except ReferenceError:
                    break
                except Exception:
                    # surface all exceptions to the main thread
                    self._worker_exception = sys.exc_info()
                    break
            try:
                self.cleanup()
            except ReferenceError as e:
                log.debug("Attempt to cleanup consumer failed")
                log.exception(e)
            log.debug("Fetcher thread exiting")
        log.info("Starting %s fetcher threads", self._num_consumer_fetchers)
        return [self._cluster.handler.spawn(fetcher, name="pykafka.SimpleConsumer.fetcher")
                for i in range(self._num_consumer_fetchers)]

    def __iter__(self):
        """Yield an infinite stream of messages until the consumer times out"""
        while True:
            message = self.consume(block=True)
            if not message:
                return
            yield message

    def consume(self, block=True, unblock_event=None):
        """Get one message from the consumer.

        :param block: Whether to block while waiting for a message
        :type block: bool
        :param unblock_event: Return when the event is set()
        :type unblock_event: :class:`threading.Event`
        """
        timeout = None
        if block:
            if self._consumer_timeout_ms > 0:
                timeout = float(self._consumer_timeout_ms) / 1000
            else:
                timeout = 1.0

        ret = None
        while True:
            self._raise_worker_exceptions()
            self._cluster.handler.sleep()
            if self._messages_arrived.acquire(blocking=block, timeout=timeout):
                # by passing through this semaphore, we know that at
                # least one message is waiting in some queue.
                if not self._running:
                    raise ConsumerStoppedException()
                message = None
                while not message:
                    owned_partition = next(self.partition_cycle)
                    message = owned_partition.consume()
                ret = message
                break
            else:
                if not self._running:
                    raise ConsumerStoppedException()
                elif not block or self._consumer_timeout_ms > 0:
                    ret = None
                    break
            if unblock_event and unblock_event.is_set():
                return

        if any(op.message_count <= self._queued_max_messages
               for op in itervalues(self._partitions)):
            if not self._slot_available.is_set():
                self._slot_available.set()

        if self._deserializer is not None:
            ret.value, ret.partition_key = self._deserializer(ret.value,
                                                              ret.partition_key)
        return ret

    def _auto_commit(self):
        """Commit offsets only if it's time to do so"""
        if not self._auto_commit_enable or self._auto_commit_interval_ms == 0:
            return

        if (time.time() - self._last_auto_commit) * 1000.0 >= self._auto_commit_interval_ms:
            log.debug("Autocommitting consumer offset for consumer group %s and topic %s",
                      self._consumer_group, self._topic.name)
            if self._consumer_group is not None:
                self.commit_offsets()
            self._last_auto_commit = time.time()

    def commit_offsets(self, partition_offsets=None):
        """Commit offsets for this consumer's partitions

        Uses the offset commit/fetch API

        :param partition_offsets: (`partition`, `offset`) pairs to
            commit where `partition` is the partition for which to commit the offset
            and `offset` is the offset to commit for the partition. Note that using
            this argument when `auto_commit_enable` is enabled can cause inconsistencies
            in committed offsets. For best results, use *either* this argument *or*
            `auto_commit_enable`.
        :type partition_offsets: Sequence of tuples of the form
            (:class:`pykafka.partition.Partition`, int)
        """
        if not self._consumer_group:
            raise Exception("consumer group must be specified to commit offsets")

        if partition_offsets is None:
            partition_offsets = [(p, None) for p in self._partitions.keys()]

        # turn Partitions into their corresponding OwnedPartitions
        try:
            owned_partition_offsets = {self._partitions[p]: offset
                                       for p, offset in partition_offsets}
        except KeyError as e:
            raise KafkaException("Unknown partition supplied to commit_offsets\n%s", e)
        reqs = [p.build_offset_commit_request(offset=o) for p, o
                in iteritems(owned_partition_offsets)]

        log.debug("Committing offsets for %d partitions to broker id %s", len(reqs),
                  self._group_coordinator.id)
        for i in range(self._offsets_commit_max_retries):
            if i > 0:
                log.debug("Retrying")
            self._cluster.handler.sleep(i * (self._offsets_channel_backoff_ms / 1000))

            try:
                response = self._group_coordinator.commit_consumer_group_offsets(
                    self._consumer_group, self._generation_id, self._consumer_id, reqs)
            except (SocketDisconnectedError, IOError):
                log.error("Error committing offsets for topic '%s' from consumer id '%s'"
                          "(SocketDisconnectedError)",
                          self._topic.name, self._consumer_id)
                if i >= self._offsets_commit_max_retries - 1:
                    raise
                self._update()
                continue

            parts_by_error = handle_partition_responses(
                self._default_error_handlers,
                response=response,
                partitions_by_id=self._partitions_by_id)
            if (len(parts_by_error) == 1 and 0 in parts_by_error) or \
                    len(parts_by_error) == 0:
                break
            log.error("Error committing offsets for topic '%s' from consumer id '%s'"
                      "(errors: %s)", self._topic.name, self._consumer_id,
                      {ERROR_CODES[err]: [op.partition.id for op, _ in parts]
                       for err, parts in iteritems(parts_by_error)})

            # retry only the partitions that errored
            if 0 in parts_by_error:
                parts_by_error.pop(0)
            errored_partitions = [
                op for code, err_group in iteritems(parts_by_error)
                for op, res in err_group
            ]
            reqs = [op.build_offset_commit_request(offset=owned_partition_offsets[op])
                    for op in errored_partitions]

    def fetch_offsets(self):
        """Fetch offsets for this consumer's topic

        Uses the offset commit/fetch API

        :return: List of (id, :class:`pykafka.protocol.OffsetFetchPartitionResponse`)
            tuples
        """
        if not self._consumer_group:
            raise Exception("consumer group must be specified to fetch offsets")
        if not self._partitions:
            return []

        def _handle_success(parts):
            partition_offsets_to_reset = []
            for owned_partition, pres in parts:
                # If Kafka returned -1, that means that no
                # offset was associated with this consumer group.
                # This partition will have its offset reset.
                if pres.offset == -1:
                    log.debug("Partition %s has no committed offsets in "
                              "consumer group %s.  Resetting to %s",
                              owned_partition.partition.id,
                              self._consumer_group,
                              self._auto_offset_reset)
                    partition_offsets_to_reset.append((
                        owned_partition.partition,
                        self._auto_offset_reset
                    ))
                else:
                    log.debug("Set offset for partition %s to %s",
                              owned_partition.partition.id,
                              pres.offset - 1)
                    # offset fetch requests return the next offset to consume,
                    # so account for this here by passing offset - 1
                    owned_partition.set_offset(pres.offset - 1)

            # If any partitions didn't have a committed offset,
            # then reset those partition's offsets.
            if partition_offsets_to_reset:
                self.reset_offsets(partition_offsets_to_reset)

        reqs = [p.build_offset_fetch_request() for p in self._partitions.values()]
        success_responses = []

        log.debug("Fetching offsets for %d partitions from broker id %s", len(reqs),
                  self._group_coordinator.id)

        for i in range(self._offsets_fetch_max_retries):
            if i > 0:
                log.debug("Retrying offset fetch")

            res = self._group_coordinator.fetch_consumer_group_offsets(self._consumer_group, reqs)
            parts_by_error = handle_partition_responses(
                self._default_error_handlers,
                response=res,
                success_handler=_handle_success if self._reset_offset_on_fetch else None,
                partitions_by_id=self._partitions_by_id)

            success_responses.extend([(op.partition.id, r)
                                      for op, r in parts_by_error.get(0, [])])
            if len(reqs) == 0 or (len(parts_by_error) == 1 and 0 in parts_by_error):
                return success_responses
            log.error("Error fetching offsets for topic '%s' (errors: %s)",
                      self._topic.name,
                      {ERROR_CODES[err]: [op.partition.id for op, _ in parts]
                       for err, parts in iteritems(parts_by_error)})

            self._cluster.handler.sleep(i * (self._offsets_channel_backoff_ms / 1000))

            to_retry = [pair for err in itervalues(parts_by_error) for pair in err]
            reqs = [p.build_offset_fetch_request() for p, _ in to_retry]

        raise KafkaException(parts_by_error)

    def reset_offsets(self, partition_offsets=None):
        """Reset offsets for the specified partitions

        For each value provided in `partition_offsets`: if the value is an integer,
        immediately reset the partition's internal offset counter to that value. If
        it's a `datetime.datetime` instance or a valid `OffsetType`, issue a
        `ListOffsetRequest` using that timestamp value to discover the latest offset
        in the latest log segment before that timestamp, then set the partition's
        internal counter to that value.

        :param partition_offsets: (`partition`, `timestamp_or_offset`) pairs to
            reset where `partition` is the partition for which to reset the offset
            and `timestamp_or_offset` is EITHER the timestamp before which to find
            a valid offset to set the partition's counter to OR the new offset the
            partition's counter should be set to.
        :type partition_offsets: Sequence of tuples of the form
            (:class:`pykafka.partition.Partition`, int OR `datetime.datetime`)
        """
        def _handle_success(parts):
            for owned_partition, pres in parts:
                if len(pres.offset) > 0:
                    # there is at least one log segment that starts before the given
                    # timestamp. set the counter to the latest offset of the latest
                    # log segment before the given timestamp.
                    owned_partition_offsets[owned_partition] = pres.offset[0] - 1
                else:
                    log.warning("Partition {id_}: no offsets available before {offset}."
                                "Defaulting to OffsetType.EARLIEST.".format(
                                    id_=owned_partition.partition.id,
                                    offset=owned_partition_timestamps[owned_partition]))
                    owned_partition_offsets[owned_partition] = OffsetType.EARLIEST

        if partition_offsets is None:
            partition_offsets = [(a, self._auto_offset_reset)
                                 for a in self._partitions.keys()]

        try:
            owned_partition_offsets = {self._partitions[p]: offset
                                       for p, offset in partition_offsets}
        except KeyError as e:
            raise KafkaException("Unknown partition supplied to reset_offsets\n%s", e)

        log.info("Resetting offsets for %s partitions", len(list(owned_partition_offsets)))

        owned_partition_timestamps = {
            op: timestamp for op, timestamp
            in iteritems(owned_partition_offsets)
            if isinstance(timestamp, dt.datetime) or timestamp in MAGIC_OFFSETS}
        if owned_partition_timestamps:
            for i in range(self._offsets_reset_max_retries):
                by_leader = defaultdict(list)
                for partition, timestamp in iteritems(owned_partition_timestamps):
                    by_leader[partition.partition.leader].append((partition, timestamp))
                for broker, timestamps in iteritems(by_leader):
                    reqs = [owned_partition.build_offset_request(timestamp)
                            for owned_partition, timestamp in timestamps]
                    response = broker.request_offset_limits(reqs)
                    parts_by_error = handle_partition_responses(
                        self._default_error_handlers,
                        response=response,
                        success_handler=_handle_success,
                        partitions_by_id=self._partitions_by_id)
                    if 0 in parts_by_error:
                        successful = [part for part, _ in parts_by_error.pop(0)]
                        list(map(owned_partition_timestamps.pop, successful))
                    if not parts_by_error:
                        continue
                    log.error("Error in ListOffsetRequest for topic '%s' (errors: %s)",
                              self._topic.name,
                              {ERROR_CODES[err]: [op.partition.id for op, _ in parts]
                               for err, parts in iteritems(parts_by_error)})
                    self._cluster.handler.sleep(i * (self._offsets_channel_backoff_ms / 1000))
                if not owned_partition_timestamps:
                    break
                log.debug("Retrying offset request")
            if owned_partition_timestamps:
                raise OffsetRequestFailedError("Offset request failed after %d "
                                               "retries", self._offsets_reset_max_retries)

        sorted_offsets = sorted(iteritems(owned_partition_offsets),
                                key=lambda k: k[0].partition.id)
        for owned_partition, offset in sorted_offsets:
            if not isinstance(offset, int):
                raise ValueError("Invalid offset value encountered in reset_offsets:\n\t"
                                 "Partition {pid} got offset '{offset}'."
                                 .format(pid=owned_partition.partition.id, offset=offset))
            with owned_partition.fetch_lock:
                owned_partition.flush()
                owned_partition.set_offset(offset)

        if self._consumer_group is not None:
            self.commit_offsets()

    def fetch(self):
        """Fetch new messages for all partitions

        Create a FetchRequest for each broker and send it. Enqueue each of the
        returned messages in the approprate OwnedPartition.
        """
        def _handle_success(parts):
            for owned_partition, pres in parts:
                if len(pres.messages) > 0:
                    log.debug("Fetched %s messages for partition %s",
                              len(pres.messages), owned_partition.partition.id)
                    owned_partition.enqueue_messages(pres.messages)
                    log.debug("Partition %s queue holds %s messages",
                              owned_partition.partition.id,
                              owned_partition.message_count)

        def unlock_partitions(parts):
            for owned_partition in parts:
                owned_partition.fetch_lock.release()

        self._wait_for_slot_available()
        sorted_by_leader = sorted(iteritems(self._partitions_by_leader),
                                  key=lambda k: k[0].id)
        for broker, owned_partitions in sorted_by_leader:
            partition_reqs = {}
            sorted_offsets = sorted(owned_partitions, key=lambda k: k.partition.id)
            for owned_partition in sorted_offsets:
                # attempt to acquire lock, just pass if we can't
                if owned_partition.fetch_lock.acquire(False):
                    partition_reqs[owned_partition] = None
                    if owned_partition.message_count < self._queued_max_messages:
                        fetch_req = owned_partition.build_fetch_request(
                            self._fetch_message_max_bytes)
                        partition_reqs[owned_partition] = fetch_req
            if partition_reqs:
                try:
                    response = broker.fetch_messages(
                        [a for a in itervalues(partition_reqs) if a],
                        timeout=self._fetch_wait_max_ms,
                        min_bytes=self._fetch_min_bytes
                    )
                except (IOError, SocketDisconnectedError):
                    unlock_partitions(iterkeys(partition_reqs))
                    if self._running:
                        log.info("Updating cluster in response to error in fetch() "
                                 "for broker id %s", broker.id)
                        self._update()
                    # If the broker dies while we're supposed to stop,
                    # it's fine, and probably an integration test.
                    return
                except socket.error:
                    raise ValueError("Failed to decode IO buffer. Ensure that "
                                     "the KafkaClient's broker_version kwarg "
                                     "matches the version of the Kafka cluster.")
                parts_by_error = build_parts_by_error(response, self._partitions_by_id)
                handle_partition_responses(
                    self._default_error_handlers,
                    parts_by_error=parts_by_error,
                    success_handler=_handle_success)
                unlock_partitions(iterkeys(partition_reqs))

    def _wait_for_slot_available(self):
        """Block until at least one queue has less than `_queued_max_messages`"""
        if all(op.message_count >= self._queued_max_messages
               for op in itervalues(self._partitions)):
            for op in itervalues(self._partitions):
                op.fetch_lock.acquire()
            if all(op.message_count >= self._queued_max_messages
                   for op in itervalues(self._partitions)):
                self._slot_available.clear()
            for op in itervalues(self._partitions):
                op.fetch_lock.release()
            while not self._slot_available.is_set():
                self._cluster.handler.sleep()
                self._slot_available.wait(5)


class OwnedPartition(object):
    """A partition that is owned by a SimpleConsumer.

    Used to keep track of offsets and the internal message queue.
    """

    def __init__(self,
                 partition,
                 handler=None,
                 semaphore=None,
                 compacted_topic=False,
                 consumer_id=b''):
        """
        :param partition: The partition to hold
        :type partition: :class:`pykafka.partition.Partition`
        :param consumer_id: The ID of the parent consumer
        :type consumer_id: bytes
        :param handler: The :class:`pykafka.handlers.Handler` instance to use
            to generate a lock
        type handler: :class:`pykafka.handler.Handler`
        :param semaphore: A Semaphore that counts available messages and
            facilitates non-busy blocking
        :type semaphore: :class:`pykafka.utils.compat.Semaphore`
        :param compacted_topic: Set to read from a compacted topic. Forces
            consumer to use less stringent ordering logic when because compacted
            topics do not provide offsets in strict incrementing order.
        :type compacted_topic: bool
        """
        self.partition = partition
        self._consumer_id = consumer_id
        self._messages = Queue()
        self._messages_arrived = semaphore
        self._is_compacted_topic = compacted_topic
        self.last_offset_consumed = -1
        self.next_offset = 0
        self.fetch_lock = handler.RLock() if handler is not None else threading.RLock()
        self.set_consumer_id(self._consumer_id)

    def set_consumer_id(self, value):
        self._consumer_id = value
        # include consumer id in offset metadata for debugging
        self._offset_metadata = {
            'consumer_id': get_string(self._consumer_id),
            'hostname': socket.gethostname()
        }
        # precalculate json to avoid expensive operation in loops
        self._offset_metadata_json = json.dumps(self._offset_metadata)

    @property
    def message_count(self):
        """Count of messages currently in this partition's internal queue"""
        return self._messages.qsize()

    def flush(self):
        """Flush internal queue"""
        # Swap out _messages so a concurrent consume/enqueue won't interfere
        tmp = self._messages
        self._messages = Queue()
        while True:
            try:
                tmp.get_nowait()
                self._messages_arrived.acquire(blocking=False)
            except Empty:
                break
        log.info("Flushed queue for partition %d", self.partition.id)

    def set_offset(self, last_offset_consumed):
        """Set the internal offset counters

        :param last_offset_consumed: The last committed offset for this
            partition
        :type last_offset_consumed: int
        """
        self.last_offset_consumed = last_offset_consumed
        self.next_offset = last_offset_consumed + 1

    def build_offset_request(self, offsets_before):
        """Create a :class:`pykafka.protocol.PartitionOffsetRequest` for this
            partition

        :param offsets_before: Timestamp indicating the
            latest write time for returned offsets. Only offsets of messages
            written before this timestamp will be returned. Permissible
            special values are `common.OffsetType.LATEST`, indicating that
            offsets from all available log segments should be returned, and
            `common.OffsetType.EARLIEST`, indicating that only the offset of
            the earliest available message should be returned.
        :type offsets_before: `datetime.datetime`
        """
        if isinstance(offsets_before, dt.datetime):
            offsets_before = round((offsets_before - EPOCH).total_seconds() * 1000)
        elif offsets_before not in MAGIC_OFFSETS:
            raise ValueError("offsets_before is an invalid timestamp: {}"
                             .format(offsets_before))
        return PartitionOffsetRequest(self.partition.topic.name, self.partition.id,
                                      offsets_before, 1)

    def build_fetch_request(self, max_bytes):
        """Create a :class:`pykafka.protocol.FetchPartitionRequest` for this
            partition.

        :param max_bytes: The number of bytes of messages to
            attempt to fetch
        :type max_bytes: int
        """
        return PartitionFetchRequest(
            self.partition.topic.name, self.partition.id,
            self.next_offset, max_bytes)

    def build_offset_commit_request(self, offset=None):
        """Create a :class:`pykafka.protocol.PartitionOffsetCommitRequest`
            for this partition

        :param offset: The offset to send in the request. If None, defaults to
            last_offset_consumed + 1
        :type offset: int
        """
        return PartitionOffsetCommitRequest(
            self.partition.topic.name,
            self.partition.id,
            offset if offset is not None else self.last_offset_consumed + 1,
            int(time.time() * 1000),
            get_bytes('{}'.format(self._offset_metadata_json))
        )

    def build_offset_fetch_request(self):
        """Create a PartitionOffsetFetchRequest for this partition
        """
        return PartitionOffsetFetchRequest(
            self.partition.topic.name,
            self.partition.id
        )

    def consume(self):
        """Get a single message from this partition"""
        try:
            message = self._messages.get_nowait()
            self.last_offset_consumed = message.offset
            return message
        except Empty:
            return None

    def enqueue_messages(self, messages):
        """Put a set of messages into the internal message queue

        :param messages: The messages to enqueue
        :type messages: Iterable of :class:`pykafka.common.Message`
        """
        for message in messages:
            # enforce ordering of messages
            if (self._is_compacted_topic and message.offset < self.next_offset) or \
                    (not self._is_compacted_topic and message.offset != self.next_offset):
                log.debug("Skipping enqueue for offset (%s) "
                          "not equal to next_offset (%s)",
                          message.offset, self.next_offset)
                continue

            message.partition = self.partition
            if message.partition_id != self.partition.id:
                log.error("Partition %s enqueued a message meant for partition %s",
                          self.partition.id, message.partition_id)
            message.partition_id = self.partition.id
            self._messages.put(message)
            self.next_offset = message.offset + 1

            if self._messages_arrived is not None:
                self._messages_arrived.release()
